"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

This module implements functionality for tracing and logging console messages.

Classes
-------
ColoredOutputFormatter(logging.Formatter)
    A log formatter for coloring log messages based on log level.
ConsoleSpinner
    A console messages spinner context manager.
LogDispatcher
    A dispatcher for configuring and handling log messages.
LogSettings
    Dataclass containing settings for a LogDispatcher.
MessageSpinner
    A spinner for console messages.
StreamWriteProxy
    A proxy for manipulating the write methods of streams.

Attributes
----------
console_lock
    The lock to acquire for manipulating the console messages.
"""

import builtins
import contextlib
import copy
import dataclasses
import functools
import itertools
import logging
import pathlib
import re
import shutil
import sys
import threading
import time
import typing

console_lock = threading.Lock()
logger = logging.getLogger(__name__)


class ColoredOutputFormatter(logging.Formatter):
    """
    A log formatter for coloring log messages based on log level.

    Inserts ANSI escape codes to color lines based on their log level:

    - logging.DEBUG : gray
    - logging.WARNING : yellow
    - logging.ERROR : dark read
    - logging.CRITICAL : brighter red

    Attributes
    ----------
    log_level_fg_colors : dict
        Mapping of log level to ANSI escape color codes.
    reset_color : str
        The ANSI escape code to reset (color) formatting.
    """

    log_level_fg_colors = {
        # https://en.wikipedia.org/wiki/ANSI_escape_code#8-bit
        logging.DEBUG: "\x1b[38;5;8m",  # gray
        logging.INFO: "",  # no special color
        logging.WARNING: "\x1b[38;5;3m",  # yellow
        logging.ERROR: "\x1b[38;5;1m",  # dark red
        logging.CRITICAL: "\x1b[38;5;160m",  # brighter red
    }
    reset_color = "\x1b[0m"

    def format(self, record):
        """Format the specified record as colored text."""
        fg_color = self.log_level_fg_colors.get(record.levelno, "")
        if fg_color:
            color_record = copy.copy(record)
            color_record.msg = fg_color + record.msg + self.reset_color
            return super().format(color_record)
        else:
            return super().format(record)


class ConsoleSpinner:
    """
    A console messages spinner context manager.

    Creates a context in which messages to `sys.stdout` and `sys.stderr` are
    prepended with a spinner to indicate that the program is still progressing
    while we are waiting for the next message to be displayed on the console.

    Notes
    -----
    The :class:`ConsoleSpinner` is implemented by monkeypatching
    :py:data:`sys.stdout`, :py:data:`sys.stderr`, and :py:func:`input`.

    The :py:meth:`sys.stdout.write` and :py:meth:`sys.stderr.write` methods are
    replaced by the :meth:`_update_spinner_msg` method which starts a new
    :class:`~cotainr.tracing.MessageSpinner` thread every time a new message is
    written to `stdout` / `stderr`. In order to avoid an infinite recursion
    when the text is actually written to the console, `sys.stdout` and
    `sys.stderr` are wrapped by :class:`~cotainr.tracing.StreamWriteProxy`
    instances.

    The :py:func:`input` function is wrapped to make sure that the spinner is
    stopped while waiting for the user to provide their input.

    As we only ever have a single console which we are interacting with via
    stdin/stdout/stderr, it makes sense to have a single corresponding
    ConsoleSpinner that actually manipulates stdin/stdout/stderr. We keep track
    of this "main" instance by having it acquire the module level
    `console_lock`.

    Nothing is done to handle the use of the :py:mod:`readline` module. It may
    or may not work in a :class:`ConsoleSpinner` context.

    This :class:`ConsoleSpinner` class is only intended to be used with the
    :py:mod:`threading` module. It doesn't support :py:mod:`multiprocessing`.
    Using the :class:`ConsoleSpinner` within multiple threads should be
    avoided, since this may lead to strange race conditions. If unavoidable
    make sure that all those threads are started within a ConsoleSpinner
    context to ensure correct operation.
    """

    def __init__(self):
        """Construct the ConsoleSpinner context manager."""
        self._stdout_proxy = StreamWriteProxy(stream=sys.stdout)
        self._stderr_proxy = StreamWriteProxy(stream=sys.stderr)
        self._wrapped_stdout_write = functools.partial(
            self._update_spinner_msg, stream=self._stdout_proxy
        )
        self._wrapped_stderr_write = functools.partial(
            self._update_spinner_msg, stream=self._stderr_proxy
        )
        self._true_input_func = builtins.input
        self._inside_this_contextmanager = False
        self._in_nested_context_count = 0
        self._spinning_msg = None
        self._lock = threading.Lock()  # Lock for keeping track of spinning messages

    def __enter__(self):
        """
        Set up the console spinner context.

        Returns
        -------
        self : :class:`ConsoleSpinner`
            The console spinner context.
        """
        with self._lock:
            if console_lock.acquire(blocking=False):
                # Wrap stdout, stderr, and input to prepend spinner to all console
                # messages
                sys.stdout.write = self._wrapped_stdout_write
                sys.stderr.write = self._wrapped_stderr_write
                builtins.input = self._thread_safe_input(builtins.input)
            else:
                # We are already within another ConsoleSpinner context which
                # handles the console spinner - this context does nothing
                self._in_nested_context_count += 1

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Tear down the console spinner context."""
        if self._in_nested_context_count == 0:
            # This is the "outermost" ConsoleSpinner context - we need to
            # properly remove the console spinner setup
            with self._lock:
                if self._spinning_msg is not None:
                    # Stop currently spinning message
                    self._spinning_msg.stop()
                    self._spinning_msg = None

                # Restore true stdout, stderr, and input
                sys.stdout.write = self._stdout_proxy.true_stream_write
                sys.stderr.write = self._stderr_proxy.true_stream_write
                builtins.input = self._true_input_func
                console_lock.release()
        else:
            self._in_nested_context_count -= 1

    def _update_spinner_msg(self, s, /, *, stream):
        """
        Update the spinning message.

        Stops the current spinning messages thread and starts a new
        :class:`~cotainr.tracing.MessageSpinner` thread for the new message.

        Parameters
        ----------
        s : str
            The new message to prepend with a spinner.
        stream : :py:class:`io.TextIOWrapper`
            The text stream on which to spin the message.
        """
        with self._lock:
            if console_lock.locked():
                # Only start a MessageSpinner if we are still within a
                # ConsoleSpinner context. Handles the possible race condition
                # of another thread trying to update the spinning message while
                # the thread owning the ConsoleSpinner context is exiting it.
                if s.strip():
                    # Only update the spinning message if it actually contains anything
                    # This also handles the problem with `print()` making two
                    # writes to the file descriptor, one with the message and one
                    # with the `end`.
                    if self._spinning_msg is not None:
                        # Stop currently spinning message
                        self._spinning_msg.stop()

                    # Start spinning the new message
                    self._spinning_msg = MessageSpinner(msg=s, stream=stream)
                    self._spinning_msg.start()
            else:
                # In case we have left the ConsoleSpinner context, simply write
                # the message as if outside the context, since now the "true"
                # write method should have been restored.
                stream.write(s)

    def _thread_safe_input(self, input_func):
        """
        Decorate a function to make input stop spinning messages.

        Stops the spinning message prior to calling the :py:func:`input` such
        that the spinning message doesn't interfere with the user input.
        """

        @functools.wraps(input_func)
        def wrapped_input_func(*args, **kwargs):
            with self._lock:
                if self._spinning_msg is not None:
                    self._spinning_msg.stop()
                    self._spinning_msg = None

                try:
                    # Temporarily restore restore the true stdout/stderr writes
                    # to allow `input()` to write its prompt without the
                    # spinner.
                    sys.stdout.write = self._stdout_proxy.true_stream_write
                    sys.stderr.write = self._stderr_proxy.true_stream_write
                    return input_func(*args, **kwargs)
                finally:
                    sys.stdout.write = self._wrapped_stdout_write
                    sys.stderr.write = self._wrapped_stderr_write

        return wrapped_input_func


class LogDispatcher:
    """
    A dispatcher for configuring and handling log messages.

    Sets up the logging machinery and provides methods for logging to the
    relevant log handlers using the correct log level - to be used in
    subprocesses. See the :ref:`cotainr Tracing & Logging documentation page
    <tracing_logging>` for more details.

    Parameters
    ----------
    name : str
        The name to use to identify this log dispatcher in log messages.
    map_log_level_func : Callable
        A callable that maps a message to a log level to use for the message.
    log_settings : :class:`~cotainr.tracing.LogSettings`
        The settings to use when setting up the logging machinery.
    filters : list of :py:class:`logging.Filter`, optional
        The filters to add to the log handlers (the default is None with
        implies that no filters are added to the log handlers).

    Attributes
    ----------
    verbosity : int
        The cotainr verbosity level used by the log dispatcher.
    map_log_level : Callable
        The callable that maps a message to a log level to use for the message.
    log_file_path : :py:class:`pathlib.Path`, optional
        The prefix of the file path to save logs to, if specified.
    no_color : bool
        The indicator of whether or not to disable the coloring of console
        message.
    logger_stdout : :py:class:`logging.Logger`
        The logger used to log to `stdout`.
    logger_stderr : :py:class:`logging.Logger`
        The logger used to log to `stderr`.
    """

    def __init__(
        self,
        *,
        name,
        map_log_level_func,
        log_settings,
        filters=None,
    ):
        """Set up the log dispatcher."""
        self.map_log_level = map_log_level_func
        self.verbosity = log_settings.verbosity
        self.log_file_path = log_settings.log_file_path
        self.no_color = log_settings.no_color
        log_level = self._determine_log_level(verbosity=log_settings.verbosity)

        # Setup cotainr log format
        if self.verbosity >= 3:
            log_fmt = "%(asctime)s - %(name)s:-:%(levelname)s: %(message)s"
        else:
            log_fmt = "%(name)s:-: %(message)s"

        # Setup log handlers
        stdout_handlers = [logging.StreamHandler(stream=sys.stdout)]
        stderr_handlers = [logging.StreamHandler(stream=sys.stderr)]

        if self.log_file_path is not None:
            stdout_handlers.append(
                logging.FileHandler(
                    self.log_file_path.with_suffix(self.log_file_path.suffix + ".out")
                )
            )
            stderr_handlers.append(
                logging.FileHandler(
                    self.log_file_path.with_suffix(self.log_file_path.suffix + ".err")
                )
            )
        for handler in stdout_handlers + stderr_handlers:
            handler.setLevel(log_level)
            handler.setFormatter(logging.Formatter(log_fmt))
            if filters is not None:
                for filter in filters:
                    handler.addFilter(filter)

        if not self.no_color:
            # Replace console formatters with one that colors the output
            stdout_handlers[0].setFormatter(ColoredOutputFormatter(log_fmt))
            stderr_handlers[0].setFormatter(ColoredOutputFormatter(log_fmt))

        # Setup loggers
        self.logger_stdout = logging.getLogger(f"{name}.out")
        self.logger_stdout.setLevel(log_level)
        for handler in stdout_handlers:
            self.logger_stdout.addHandler(handler)

        self.logger_stderr = logging.getLogger(f"{name}.err")
        self.logger_stderr.setLevel(log_level)
        for handler in stderr_handlers:
            self.logger_stderr.addHandler(handler)

        logger.debug(
            "LogDispatcher: %s, LEVEL: %s, STDOUT handlers: %s, STDERR handlers: %s",
            name,
            log_level,
            self.logger_stdout.handlers,
            self.logger_stderr.handlers,
        )

    def log_to_stderr(self, msg):
        """
        Log a message to `stderr`.

        Determines the log level based on the `map_log_level` function and logs
        the `msg` to `stderr` using that log level.

        Parameters
        ----------
        msg : str
            The message to log.
        """
        self.logger_stderr.log(level=self.map_log_level(msg), msg=msg)

    def log_to_stdout(self, msg):
        """
        Log a message to `stdout`.

        Determines the log level based on the `map_log_level` function and logs
        the `msg` to `stdout` using that log level.

        Parameters
        ----------
        msg : str
            The message to log.
        """
        self.logger_stdout.log(level=self.map_log_level(msg), msg=msg)

    @contextlib.contextmanager
    def prefix_stderr_name(self, *, prefix):
        """
        Manage a context to prefix the `stderr` logger name.

        When inside the context, the name of the `stderr` logger is changed to
        be prefixed by "`prefix`/".

        Parameters
        ----------
        prefix : str
            The prefix add to the `stderr` logger name.
        """
        logger_stderr_name = self.logger_stderr.name
        self.logger_stderr.name = prefix + "/" + logger_stderr_name
        yield
        self.logger_stderr.name = logger_stderr_name

    @staticmethod
    def _determine_log_level(*, verbosity):
        """
        Map cotainr verbosity level to a logger and handler log level.

        Parameters
        ----------
        verbosity : int
            The cotainr verbosity level used by the log dispatcher.

        Returns
        -------
        log_level : int
            One of the standard log levels (DEBUG, INFO, WARNING, ERROR, or
            CRITICAL).
        """
        if verbosity <= -1:
            log_level = logging.CRITICAL
        elif verbosity == 0:
            log_level = logging.WARNING
        elif verbosity == 1 or verbosity == 2:
            log_level = logging.INFO
        elif verbosity >= 3:
            log_level = logging.DEBUG
        else:
            # This should not happen, but if you somehow specify e.g.
            # verbosity=numpy.NaN, you will end up here...
            raise ValueError(
                f"Somehow we ended up with a {verbosity=} of {type(verbosity)=} "
                "that does not compare well with integers."
            )

        return log_level


@dataclasses.dataclass
class LogSettings:
    """
    Dataclass containing settings for a LogDispatcher.

    Attributes
    ----------
    verbosity : int, default=0
        The cotainr verbosity level used by the log dispatcher.
    log_file_path : :py:class:`pathlib.Path`, default=None
        The prefix of the file path to save logs to, if specified.
    no_color : bool, default=False
        The indicator of whether or not to disable the coloring of console
        message.
    """

    verbosity: int = 0
    log_file_path: typing.Optional[pathlib.Path] = None
    no_color: bool = False

    def __post_init__(self):
        """Cast fields to their expected types."""
        self.verbosity = int(self.verbosity)
        if self.log_file_path is not None:
            self.log_file_path = pathlib.Path(self.log_file_path)
        self.no_color = bool(self.no_color)


class MessageSpinner:
    """
    A spinner for console messages.

    Creates a thread that continuously updates a prepended spinner to `msg` on
    `stream`. While the spinner is active, the `msg` is truncated to fit within
    a single line. When the spinner is stopped, the full `msg` (without
    spinner) is written to the `stream`.

    Parameters
    ----------
    msg : str
        The message to prepend with a spinner.
    stream : :py:class:`io.TextIOWrapper`
        The text stream on which to spin the message.

    Notes
    -----
    The spinner is added by continuously rewriting the current console line to
    update the spinning character prepended to `msg`. This is done using CR and
    ANSI escape codes. To avoid interfering with our manipulation of the
    console line, trailing newlines and ANSI codes, except "Select Graphics
    Rendition (SGR)" codes used for coloring console lines, are stripped from
    `msg` before it is displayed on the console.
    """

    def __init__(self, *, msg, stream):
        """Construct the message spinner."""
        self._msg = msg
        self._stream = stream

        self._spinner_cycle = itertools.cycle("⣾⣷⣯⣟⡿⢿⣻⣽")
        self._spinner_thread = threading.Thread(target=self._spin_msg)
        self._spinner_sleep_interval = 0.1
        self._spinner_delay_time = 0.025
        self._stop_signal = threading.Event()
        self._print_width = (
            # account for leading spinner + whitespace (2 chars)
            # and trailing dots (3 chars)
            shutil.get_terminal_size()[0] - 5
        )
        self._ansi_escape_re = re.compile(
            # Based on r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])" from
            # https://stackoverflow.com/a/14693789 with all "Select Graphics
            # Rendition (SGR)" codes (those ending with "m") passed through.
            # See also: https://notes.burke.libbey.me/ansi-escape-codes/
            r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*([@-l]|[n-~]))"
        )
        self._newline_at_end_re = re.compile(
            # Find newlines at the end of a string even if the string is
            # wrapped in a set of SGR codes.
            r"\n+(?=(?:\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*m))+$|$)"
        )
        self._running = False

        # \033[2K erases the old line to avoid the extra two characters at the
        # end of the line stemming from the shift of the line due to the
        # leading spinner character and a whitespace
        self._clear_line_code = "\r" + "\x1b[2K"
        self._reset_SGR = "\x1b[0m"

    def start(self):
        """Start the message spinner thread."""
        self._spinner_thread.start()
        self._running = True

    def stop(self):
        """Stop the message spinner thread."""
        if self._running:
            self._stop_signal.set()
            self._spinner_thread.join()
            self._running = False

    def _spin_msg(self):
        """
        Spin the message.

        This is the method that the thread is running to continuously update
        the spinner.
        """
        # Delay spinning a bit to avoid flaky message updates when new messages
        # arrive promptly
        time.sleep(self._spinner_delay_time)

        # Strip any newlines and ANSI escape codes that may interfere with
        # our manipulation of the console (not SGRs, though)
        msg = self._ansi_escape_re.sub("", self._msg)
        msg = self._newline_at_end_re.sub("", msg)

        # Construct a messages that is guaranteed to fit on one line with the spinner
        one_line_msg = (
            # Make room for the spinner and 3 trailing dots and make sure to
            # keep the "reset SGR" code if its present
            msg[: (self._print_width - len(self._reset_SGR))] + self._reset_SGR
            if msg.endswith(self._reset_SGR)
            else msg[: self._print_width]
        ) + "..."

        # Start spinning
        while not self._stop_signal.is_set():
            # Update spinner
            print(
                f"{self._clear_line_code}{next(self._spinner_cycle)} "  # spinner
                f"{one_line_msg}",  # possibly truncated message
                end="",  # keep overwriting current line
                file=self._stream,
            )
            time.sleep(self._spinner_sleep_interval)

        # Print message without spinner (incl. trailing newline)
        print(f"{self._clear_line_code}{msg}", file=self._stream)


class StreamWriteProxy:
    """
    A proxy for manipulating the write methods of streams.

    Provides a `write` method that guarantees to use the true `stream.write`
    method in case `stream.write` is later monkeypatched. All other method
    calls to this class are delegated to the corresponding (potentially
    monkeypatched) methods implemented by `stream`.

    Parameters
    ----------
    stream : :py:class:`io.TextIOWrapper`
        The text stream to provide a proxy for.

    Notes
    -----
    This provides a way for keeping references to the true
    :py:meth:`sys.stdout.write` and :py:meth:`sys.stderr.write` methods. This
    is useful in code that monkeypatches these methods since it is not possible
    in Python to copy :py:data:`sys.stdout` and :py:data:`sys.stderr` such that
    they can later be restored.
    """

    def __init__(self, *, stream):
        """Construct the stream write proxy."""
        self._stream = stream
        self.true_stream_write = stream.write

    def write(self, s, /):
        """
        Write a string to the stream using the true `stream.write` method.

        Parameters
        ----------
        s : str
            The string to write.

        Returns
        -------
        int
            The number of characters written.
        """
        return self.true_stream_write(s)

    def __getattr__(self, name):
        """Delegate all other attribute and method calls."""
        return getattr(self._stream, name)



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import cotainr.cli

from .stubs import StubDummyCLI


class TestMain:
    def test_main(self, capsys, monkeypatch):
        monkeypatch.setattr(cotainr.cli, "CotainrCLI", StubDummyCLI)
        cotainr.cli.main()
        cli_init, subcommand_run = capsys.readouterr().out.strip().split("\n")
        assert cli_init == "Initialized DummyCLI."
        assert subcommand_run == "Executed dummy subcommand."



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

This module implements the interaction with the container runtime.

Classes
-------
SingularitySandbox
    A Singularity container sandbox context manager.
"""

import json
import logging
import os
from pathlib import Path
import shlex
import subprocess
import sys
from tempfile import TemporaryDirectory

from . import __version__ as _cotainr_version
from . import tracing, util

logger = logging.getLogger(__name__)


class SingularitySandbox:
    """
    A Singularity container sandbox context manager.

    This creates and manipulates a `Singularity sandbox
    <http://apptainer.org/docs/user/main/build_a_container.html#creating-writable-sandbox-directories>`_,
    i.e. a temporary directory representing the container. As a final step, the
    sandbox should be converted into a SIF container image file.

    Parameters
    ----------
    base_image : str
        Base image to use for the container which may be any valid
        Apptainer/Singularity <BUILD SPEC>.
    log_settings : :class:`~cotainr.tracing.LogSettings`, optional
        The data used to setup the logging machinery (the default is None which
        implies that the logging machinery is not used).

    Attributes
    ----------
    base_image : str
        Base image to use for the container.
    sandbox_dir : :class:`os.PathLike` or None
        The path to the temporary directory containing the sandbox if within a
        sandbox context, otherwise it is None.
    log_dispatcher : :class:`~cotainr.tracing.LogDispatcher` or None.
        The log dispatcher used to process stdout/stderr message from
        Singularity commands that run in sandbox, if the logging machinery is
        used.
    architecture : str or None.
        The machine architecture of the sandbox as returned by `uname -m`. Its
        value is `None` (unknown) until entering the the sandbox context.
    """

    def __init__(self, *, base_image, log_settings=None):
        """Construct the SingularitySandbox context manager."""
        self.base_image = base_image
        self.sandbox_dir = None
        self.architecture = None
        if log_settings is not None:
            self._verbosity = log_settings.verbosity
            self.log_dispatcher = tracing.LogDispatcher(
                name=__class__.__name__,
                map_log_level_func=self._map_log_level,
                log_settings=log_settings,
            )
        else:
            self._verbosity = 0
            self.log_dispatcher = None

    def __enter__(self):
        """
        Build and enter sandbox context.

        Returns
        -------
        self : :class:`SingularitySandbox`
            The sandbox context.
        """
        # Store current directory
        self._origin = Path().resolve()

        # Create sandbox
        self._tmp_dir = TemporaryDirectory()
        self.sandbox_dir = Path(self._tmp_dir.name) / "singularity_sandbox"
        self.sandbox_dir.mkdir(exist_ok=False)
        self._subprocess_runner(
            args=self._add_verbosity_arg(
                args=[
                    "singularity",
                    "--nocolor",
                    "build",
                    "--force",  # sandbox_dir.mkdir() checks for existing sandbox image
                    "--sandbox",
                    "--fix-perms",
                    self.sandbox_dir,
                    self.base_image,
                ]
            ),
        )

        # Change directory to the sandbox
        os.chdir(self.sandbox_dir)

        # Get the architecture of the sandbox if it is not already set
        # (should not be set in real world scenarios)
        if self.architecture is None:
            arch_process = self.run_command_in_container(cmd="uname -m")
            self.architecture = arch_process.stdout.strip()

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit and destroy sandbox context."""
        os.chdir(self._origin)
        self._tmp_dir.cleanup()
        self.sandbox_dir = None

    def add_metadata(self):
        """
        Add metadata to the container sandbox.

        The following metadata is added to the container sandbox:
          - "cotainr.command": The full command line used to build the container.
          - "cotainr.version": The version of cotainr used to build the container.
          - "cotainr.url": The cotainr project url.

        The container metadata may be inspected by running `singularity inspect` on the
        built container image file.

        Notes
        -----
        The metadata entries are added to the `.singularity.d/labels.json
        <https://apptainer.org/docs/user/main/environment_and_metadata.html#singularity-d-directory>`_
        file.

        """
        self._assert_within_sandbox_context()

        labels_path = self.sandbox_dir / ".singularity.d/labels.json"
        with open(labels_path, "r+") as f:
            metadata = json.load(f)
            metadata["cotainr.command"] = " ".join(sys.argv)
            metadata["cotainr.version"] = _cotainr_version
            metadata["cotainr.url"] = "https://github.com/DeiC-HPC/cotainr"
            f.seek(0)
            json.dump(metadata, f)

    def add_to_env(self, *, shell_script):
        """
        Add `shell_script` to the sourced environment in the container.

        Parameters
        ----------
        shell_script : str
            The shell script to add to the sourced environment in the
            container.
        """
        self._assert_within_sandbox_context()

        env_file = self.sandbox_dir / ".singularity.d/env/92-cotainr-env.sh"
        if not env_file.exists():
            self._create_file(f=env_file)

        with env_file.open(mode="a") as f:
            f.write(shell_script + "\n")

    def build_image(self, *, path):
        """
        Build a SIF image file from sandbox.

        Takes the current content of the sandbox and builds a SIF container
        image from it. The container image is outputted to `path`.

        Parameters
        ----------
        path : :class:`os.PathLike`
            Path to the built container image.
        """
        self._assert_within_sandbox_context()

        self._subprocess_runner(
            args=self._add_verbosity_arg(
                args=[
                    "singularity",
                    "--nocolor",
                    "build",
                    "--force",
                    path,
                    self.sandbox_dir,
                ]
            ),
        )

    def run_command_in_container(self, *, cmd, custom_log_dispatcher=None):
        """
        Run a command in the container sandbox.

        Wraps `singularity exec` of the `cmd` in the container sandbox`
        allowing for running commands inside the container sandbox context,
        e.g. for installing software in the container sandbox.

        Parameters
        ----------
        cmd : str
            The command to run in the container sandbox.
        custom_log_dispatcher : :class:`~cotainr.tracing.LogDispatcher`, optional
            The custom log dispatcher to use when running the command (the
            default is None which implies that the `SingularitySandbox` log
            dispatcher is used).

        Returns
        -------
        process : :class:`subprocess.CompletedProcess`
            Information about the process that ran in the container sandbox.

        Notes
        -----
        We pass several flags to the `singularity exec` command to provide
        maximum compatibility with different HPC systems. In particular, we
        use:

        - `--no-home` as trying to mount the home folder on some systems (e.g.
          LUMI) causes problems. Thus, when running a command in the container,
          you cannot reference files in your home directory. Instead you must
          copy all files into the container sandbox and then reference the
          files relative to the container root.
        - `--no-umask` as some systems use a default umask (e.g. 0007 on LUMI)
          that prevents you from accessing any files added to the container as
          a regular user when you run the built container, e.g. such files are
          owned by root:root with 660 permissions for a 0007 umask. Thus, all
          files added to the container by running a command in the container
          will have file permissions 644 (Apptainer/Singularity forces the
          umask to 0022). If you need other file permissions, you must manually
          change them.
        """
        self._assert_within_sandbox_context()

        try:
            process = self._subprocess_runner(
                custom_log_dispatcher=custom_log_dispatcher,
                args=self._add_verbosity_arg(
                    args=[
                        "singularity",
                        "--nocolor",
                        "exec",
                        "--writable",
                        "--no-home",
                        "--no-umask",
                        self.sandbox_dir,
                        *shlex.split(cmd),
                    ]
                ),
            )
        except subprocess.CalledProcessError as e:
            singularity_fatal_error = "\n".join(
                [line for line in e.stderr.split("\n") if line.startswith("FATAL")]
            )
            raise ValueError(
                f"Invalid command {cmd=} passed to Singularity "
                f"resulted in the FATAL error: {singularity_fatal_error}"
            ) from e

        return process

    def _assert_within_sandbox_context(self):
        """Raise a ValueError if we are not inside the sandbox context."""
        if self.sandbox_dir is None:
            raise ValueError("The operation is only valid inside a sandbox context.")

    def _add_verbosity_arg(self, *, args):
        """
        Add a verbosity level to Singularity commands.

        A mapping of the internal cotainr verbosity level to `Singularity
        verbosity flags
        <https://apptainer.org/docs/user/main/cli/apptainer.html#options>`_.

        Parameters
        ----------
        args : list
            The list of command line arguments constituting the full
            singularity command.

        Returns
        -------
        args : list
            The updated list of singularity command line arguments.
        """
        if self._verbosity < 0:
            # --silent (-s)
            args.insert(1, "-s")
        elif self._verbosity == 0:
            # --quiet (-q)
            args.insert(1, "-q")
        elif self._verbosity == 3:
            # --verbose (-v); limited debug information
            args.insert(1, "-v")
        elif self._verbosity >= 4:
            # --debug (-d); all debug information
            args.insert(1, "-d")

        return args

    def _create_file(self, *, f):
        """
        Create any file `f` in an existing folder in the Singularity container.

        The file permissions will ignore the system umask.

        Parameters
        ----------
        f : :class:`pathlib.PosixPath`
            For example, Path("sandbox_dir/.singularity.d/env/92-cotainr-env.sh")
        """
        self._assert_within_sandbox_context()

        # ensure that the file is created *within* the container to get correct permissions, etc.
        self.run_command_in_container(cmd=f"touch {f}")

        if not f.exists():
            raise FileNotFoundError(f"Creating file {f} failed.")

    def _subprocess_runner(self, *, custom_log_dispatcher=None, args, **kwargs):
        """
        Wrap the choice of subprocess runner.

        Parameters
        ----------
        custom_log_dispatcher : :class:`~cotainr.tracing.LogDispatcher`, optional
            The custom log dispatcher to use when running the command (the
            default is None which implies that the `SingularitySandbox` log
            dispatcher is used).
        args : list or str
            Subprocess arguments.
        kwargs : dict
            Keyword arguments passed to the subprocess runner.

        Returns
        -------
        process : :class:`subprocess.CompletedProcess`
            Information about the process that ran in the container sandbox.
        """
        if custom_log_dispatcher is not None:
            # When the command to be run in the container provides its own
            # log_dispatcher
            with custom_log_dispatcher.prefix_stderr_name(
                prefix=self.__class__.__name__
            ):
                return util.stream_subprocess(
                    log_dispatcher=custom_log_dispatcher, args=args, **kwargs
                )
        else:
            # Use the SingularitySandbox log_dispatcher
            return util.stream_subprocess(
                log_dispatcher=self.log_dispatcher, args=args, **kwargs
            )

    @staticmethod
    def _map_log_level(msg):
        """
        Attempt to infer log level for a message.

        Parameters
        ----------
        msg : str
            The message to infer log level for.

        Returns
        -------
        log_level : int
            One of the standard log levels (DEBUG, INFO, WARNING, ERROR, or
            CRITICAL).
        """
        if msg.startswith("DEBUG") or msg.startswith("VERBOSE"):
            return logging.DEBUG
        elif msg.startswith("INFO") or msg.startswith("LOG"):
            return logging.INFO
        elif msg.startswith("WARNING"):
            return logging.WARNING
        elif msg.startswith("ERROR"):
            return logging.ERROR
        elif msg.startswith("ABRT") or msg.startswith("FATAL"):
            return logging.CRITICAL
        else:
            # If no prefix on message, assume its INFO level
            return logging.INFO



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import contextlib

import pytest

import cotainr.tracing

from .stubs import FixedNumberOfSpinsEvent


@pytest.fixture
def patch_disable_console_spinner(monkeypatch):
    """
    Disable the ConsoleSpinner() context.

    Replaces the console spinner context with a no-op context.
    """

    @contextlib.contextmanager
    def mock_console_spinner():
        yield

    monkeypatch.setattr(cotainr.tracing, "ConsoleSpinner", mock_console_spinner)


@pytest.fixture
def patch_fix_number_of_message_spins(spins, monkeypatch):
    """
    Force `cotainr.tracing.MessageSpinner` to update the spinning message
    exactly `spins` times before stopping.

    The `spin` parameter must be passed to the fixture using a pytest
    parameterization.
    """

    class FixedNumberOfSpinsMessageSpinner(cotainr.tracing.MessageSpinner):
        def __init__(self, *, msg, stream):
            super().__init__(msg=msg, stream=stream)
            self._stop_signal = FixedNumberOfSpinsEvent(spins=spins)

    monkeypatch.setattr(
        cotainr.tracing, "MessageSpinner", FixedNumberOfSpinsMessageSpinner
    )



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import itertools
import logging
import re

import pytest

from cotainr.tracing import LogDispatcher, LogSettings

from .data import (
    data_log_dispatcher_critical_color_log_messages,
    data_log_dispatcher_debug_color_log_messages,
    data_log_dispatcher_info_color_log_messages,
    data_log_dispatcher_info_no_color_log_messages,
    data_log_dispatcher_warning_color_log_messages,
)
from .stubs import AlwaysCompareFalse


class TestConstructor:
    def test_attributes(self):
        def loc_level_func(msg):
            pass

        log_settings = LogSettings()
        log_dispatcher = LogDispatcher(
            name="test_dispatcher_6021",
            map_log_level_func=loc_level_func,
            log_settings=log_settings,
        )
        assert log_dispatcher.map_log_level is loc_level_func
        assert log_dispatcher.verbosity == log_settings.verbosity
        assert log_dispatcher.log_file_path == log_settings.log_file_path
        assert log_dispatcher.no_color == log_settings.no_color
        assert log_dispatcher.logger_stdout.name == "test_dispatcher_6021.out"

        assert log_dispatcher.logger_stderr.name == "test_dispatcher_6021.err"

    def test_adding_filters(self):
        # Define test filters
        class TestFilter1(logging.Filter):
            pass

        class TestFilter2(logging.Filter):
            pass

        filters = [TestFilter1(), TestFilter2()]

        # Setup the LogDispatcher
        log_dispatcher = LogDispatcher(
            name="test_dispatcher_6021",
            map_log_level_func=lambda msg: 0,  # not used in test since we log directly to loggers
            log_settings=LogSettings(),
            filters=filters,
        )
        for logger in [log_dispatcher.logger_stdout, log_dispatcher.logger_stderr]:
            for handler in logger.handlers:
                assert len(handler.filters) == 2
                for handler_filter, test_filter in zip(handler.filters, filters):
                    assert handler_filter is test_filter

    @pytest.mark.parametrize("verbosity", [-1, -2, -3, -5, -1000])
    def test_log_dispatcher_critical_logging(
        self,
        verbosity,
        capsys,
        data_log_dispatcher_critical_color_log_messages,
    ):
        (
            logger_name,
            log_level_msgs,
            stdout_msgs,
            stderr_msgs,
        ) = data_log_dispatcher_critical_color_log_messages

        # Setup the LogDispatcher
        log_dispatcher = LogDispatcher(
            name=logger_name,
            map_log_level_func=lambda msg: 0,  # not used in test since we log directly to loggers
            log_settings=LogSettings(
                verbosity=verbosity, log_file_path=None, no_color=False
            ),
        )
        assert f"{logger_name}.out" in logging.Logger.manager.loggerDict
        assert f"{logger_name}.err" in logging.Logger.manager.loggerDict

        # Log test messages to log dispatcher loggers
        for logger in [log_dispatcher.logger_stdout, log_dispatcher.logger_stderr]:
            # Log messages
            for level, msg in log_level_msgs.items():
                logger.log(level=level, msg=msg)

            # Check correct log level
            assert logger.getEffectiveLevel() == logging.CRITICAL

            # Check correct handles
            # Pytest manipulates the handler streams to capture the logging output
            assert len(logger.handlers) == 1
            assert isinstance(logger.handlers[0], logging.StreamHandler)

        # Check correct logging, incl. message format, coloring, log level, and output stream
        # readouterr clears its content when returning
        stdout, stderr = capsys.readouterr()
        assert stdout.rstrip("\n").split("\n") == stdout_msgs
        assert stderr.rstrip("\n").split("\n") == stderr_msgs

    @pytest.mark.parametrize("verbosity", [3, 4, 5, 1000])
    def test_log_dispatcher_debug_logging(
        self,
        verbosity,
        capsys,
        data_log_dispatcher_debug_color_log_messages,
    ):
        (
            logger_name,
            log_level_msgs,
            expected_stdout_msgs,
            expected_stderr_msgs,
        ) = data_log_dispatcher_debug_color_log_messages

        # Setup the LogDispatcher
        log_dispatcher = LogDispatcher(
            name=logger_name,
            map_log_level_func=lambda msg: 0,  # not used in test since we log directly to loggers
            log_settings=LogSettings(
                verbosity=verbosity, log_file_path=None, no_color=False
            ),
        )
        assert f"{logger_name}.out" in logging.Logger.manager.loggerDict
        assert f"{logger_name}.err" in logging.Logger.manager.loggerDict

        # Log test messages to log dispatcher loggers
        for logger in [log_dispatcher.logger_stdout, log_dispatcher.logger_stderr]:
            # Log messages
            for level, msg in log_level_msgs.items():
                logger.log(level=level, msg=msg)

            # Check correct log level
            assert logger.getEffectiveLevel() == logging.DEBUG

            # Check correct handles
            # Pytest manipulates the handler streams to capture the logging output
            assert len(logger.handlers) == 1
            assert isinstance(logger.handlers[0], logging.StreamHandler)

        # Check correct logging, incl. message format, coloring, log level, and output stream
        # readouterr clears its content when returning
        stdout, stderr = capsys.readouterr()
        actual_stdout_msgs = stdout.rstrip("\n").split("\n")
        actual_stderr_msgs = stderr.rstrip("\n").split("\n")
        assert len(actual_stdout_msgs) == len(expected_stdout_msgs)
        assert len(actual_stderr_msgs) == len(expected_stderr_msgs)
        debug_msg_start_re = re.compile(
            # Verbose debug msg prefix along lines of:
            # 2023-09-27 10:39:07,763 -
            r"^\d{4}(-\d{2}){2} (\d{2}:){2}\d{2},\d{3} - "
        )
        for actual_msg, expected_msg in itertools.chain(
            zip(actual_stdout_msgs, expected_stdout_msgs),
            zip(actual_stderr_msgs, expected_stderr_msgs),
        ):
            assert debug_msg_start_re.match(actual_msg)
            assert actual_msg.endswith(expected_msg)

    @pytest.mark.parametrize("verbosity", [1, 2])
    def test_log_dispatcher_info_logging(
        self,
        verbosity,
        capsys,
        data_log_dispatcher_info_color_log_messages,
    ):
        (
            logger_name,
            log_level_msgs,
            stdout_msgs,
            stderr_msgs,
        ) = data_log_dispatcher_info_color_log_messages

        # Setup the LogDispatcher
        log_dispatcher = LogDispatcher(
            name=logger_name,
            map_log_level_func=lambda msg: 0,  # not used in test since we log directly to loggers
            log_settings=LogSettings(
                verbosity=verbosity, log_file_path=None, no_color=False
            ),
        )
        assert f"{logger_name}.out" in logging.Logger.manager.loggerDict
        assert f"{logger_name}.err" in logging.Logger.manager.loggerDict

        # Log test messages to log dispatcher loggers
        for logger in [log_dispatcher.logger_stdout, log_dispatcher.logger_stderr]:
            # Log messages
            for level, msg in log_level_msgs.items():
                logger.log(level=level, msg=msg)

            # Check correct log level
            assert logger.getEffectiveLevel() == logging.INFO

            # Check correct handles
            # Pytest manipulates the handler streams to capture the logging output
            assert len(logger.handlers) == 1
            assert isinstance(logger.handlers[0], logging.StreamHandler)

        # Check correct logging, incl. message format, coloring, log level, and output stream
        # readouterr clears its content when returning
        stdout, stderr = capsys.readouterr()
        assert stdout.rstrip("\n").split("\n") == stdout_msgs
        assert stderr.rstrip("\n").split("\n") == stderr_msgs

    def test_log_dispatcher_warning_logging(
        self,
        capsys,
        data_log_dispatcher_warning_color_log_messages,
    ):
        verbosity = 0
        (
            logger_name,
            log_level_msgs,
            stdout_msgs,
            stderr_msgs,
        ) = data_log_dispatcher_warning_color_log_messages

        # Setup the LogDispatcher
        log_dispatcher = LogDispatcher(
            name=logger_name,
            map_log_level_func=lambda msg: 0,  # not used in test since we log directly to loggers
            log_settings=LogSettings(
                verbosity=verbosity, log_file_path=None, no_color=False
            ),
        )
        assert f"{logger_name}.out" in logging.Logger.manager.loggerDict
        assert f"{logger_name}.err" in logging.Logger.manager.loggerDict

        # Log test messages to log dispatcher loggers
        for logger in [log_dispatcher.logger_stdout, log_dispatcher.logger_stderr]:
            # Log messages
            for level, msg in log_level_msgs.items():
                logger.log(level=level, msg=msg)

            # Check correct log level
            assert logger.getEffectiveLevel() == logging.WARNING

            # Check correct handles
            # Pytest manipulates the handler streams to capture the logging output
            assert len(logger.handlers) == 1
            assert isinstance(logger.handlers[0], logging.StreamHandler)

        # Check correct logging, incl. message format, coloring, log level, and output stream
        # readouterr clears its content when returning
        stdout, stderr = capsys.readouterr()
        assert stdout.rstrip("\n").split("\n") == stdout_msgs
        assert stderr.rstrip("\n").split("\n") == stderr_msgs

    @pytest.mark.parametrize("no_color", [True, False])
    def test_log_to_file(
        self,
        no_color,
        tmp_path,
        data_log_dispatcher_info_no_color_log_messages,
    ):
        (
            logger_name,
            log_level_msgs,
            stdout_msgs,
            stderr_msgs,
        ) = data_log_dispatcher_info_no_color_log_messages

        # Setup the LogDispatcher
        log_file_path = tmp_path / "cotainr_log"
        log_dispatcher = LogDispatcher(
            name=logger_name,
            map_log_level_func=lambda msg: 0,  # not used in test since we log directly to loggers
            log_settings=LogSettings(
                verbosity=1, log_file_path=log_file_path, no_color=no_color
            ),
        )
        assert f"{logger_name}.out" in logging.Logger.manager.loggerDict
        assert f"{logger_name}.err" in logging.Logger.manager.loggerDict

        # Log test messages to log dispatcher loggers
        for logger in [log_dispatcher.logger_stdout, log_dispatcher.logger_stderr]:
            # Log messages
            for level, msg in log_level_msgs.items():
                logger.log(level=level, msg=msg)

            # Check correct log level
            assert logger.getEffectiveLevel() == logging.INFO

            # Check correct handles
            # Pytest manipulates the handler streams to capture the logging output
            assert len(logger.handlers) == 2
            assert isinstance(logger.handlers[0], logging.StreamHandler)
            assert isinstance(logger.handlers[1], logging.FileHandler)

        # Check correct logging, incl. message format, coloring, log level to file
        assert (
            log_file_path.with_suffix(".out").read_text().rstrip("\n").split("\n")
            == stdout_msgs
        )
        assert (
            log_file_path.with_suffix(".err").read_text().rstrip("\n").split("\n")
            == stderr_msgs
        )

    def test_no_color_on_console(
        self,
        capsys,
        data_log_dispatcher_info_no_color_log_messages,
    ):
        (
            logger_name,
            log_level_msgs,
            stdout_msgs,
            stderr_msgs,
        ) = data_log_dispatcher_info_no_color_log_messages

        # Setup the LogDispatcher
        log_dispatcher = LogDispatcher(
            name=logger_name,
            map_log_level_func=lambda msg: 0,  # not used in test since we log directly to loggers
            log_settings=LogSettings(verbosity=1, log_file_path=None, no_color=True),
        )
        assert f"{logger_name}.out" in logging.Logger.manager.loggerDict
        assert f"{logger_name}.err" in logging.Logger.manager.loggerDict

        # Log test messages to log dispatcher loggers
        for logger in [log_dispatcher.logger_stdout, log_dispatcher.logger_stderr]:
            # Log messages
            for level, msg in log_level_msgs.items():
                logger.log(level=level, msg=msg)

            # Check correct log level
            assert logger.getEffectiveLevel() == logging.INFO

            # Check correct handles
            # Pytest manipulates the handler streams to capture the logging output
            assert len(logger.handlers) == 1
            assert isinstance(logger.handlers[0], logging.StreamHandler)

        # Check correct logging, incl. message format, coloring, log level, and output stream
        # readouterr clears its content when returning
        stdout, stderr = capsys.readouterr()
        assert stdout.rstrip("\n").split("\n") == stdout_msgs
        assert stderr.rstrip("\n").split("\n") == stderr_msgs


class TestLogToStderr:
    @pytest.mark.parametrize(
        "level",
        [logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
    )
    def test_log_to_correct_level(self, level, caplog):
        log_dispatcher = LogDispatcher(
            name="test_dispatcher_6021",
            map_log_level_func=lambda msg: level,
            log_settings=LogSettings(verbosity=3, log_file_path=None, no_color=True),
        )
        caplog.clear()  # Clear any logs from LogDispatcher initialization
        log_dispatcher.log_to_stderr("test_6021")
        assert len(caplog.records) == 1
        assert caplog.records[0].levelno == level
        assert caplog.records[0].name == "test_dispatcher_6021.err"
        assert caplog.records[0].msg == "test_6021"

    def test_not_stripping_whitespace(self, caplog):
        log_dispatcher = LogDispatcher(
            name="test_dispatcher_6021",
            map_log_level_func=lambda msg: logging.INFO,
            log_settings=LogSettings(verbosity=1, log_file_path=None, no_color=True),
        )
        caplog.clear()  # Clear any logs from LogDispatcher initialization
        msgs = ["  before", "after  ", "  before and after  "]
        for msg in msgs:
            log_dispatcher.log_to_stderr(msg=msg)

        assert len(caplog.records) == 3
        assert all(rec.name == "test_dispatcher_6021.err" for rec in caplog.records)
        assert all(msg == rec.msg for msg, rec in zip(msgs, caplog.records))


class TestLogToStdout:
    @pytest.mark.parametrize(
        "level",
        [logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
    )
    def test_log_to_correct_level(self, level, caplog):
        log_dispatcher = LogDispatcher(
            name="test_dispatcher_6021",
            map_log_level_func=lambda msg: level,
            log_settings=LogSettings(verbosity=3, log_file_path=None, no_color=True),
        )
        caplog.clear()  # Clear any logs from LogDispatcher initialization
        log_dispatcher.log_to_stdout("test_6021")
        assert len(caplog.records) == 1
        assert caplog.records[0].levelno == level
        assert caplog.records[0].name == "test_dispatcher_6021.out"
        assert caplog.records[0].msg == "test_6021"

    def test_not_stripping_whitespace(self, caplog):
        log_dispatcher = LogDispatcher(
            name="test_dispatcher_6021",
            map_log_level_func=lambda msg: logging.INFO,
            log_settings=LogSettings(verbosity=1, log_file_path=None, no_color=True),
        )
        caplog.clear()  # Clear any logs from LogDispatcher initialization
        msgs = ["  before", "after  ", "  before and after  "]
        for msg in msgs:
            log_dispatcher.log_to_stdout(msg=msg)

        assert len(caplog.records) == 3
        assert all(rec.name == "test_dispatcher_6021.out" for rec in caplog.records)
        assert all(msg == rec.msg for msg, rec in zip(msgs, caplog.records))


class TestPrefixStderrName:
    def test_context_prefix_changes(self, caplog):
        log_dispatcher = LogDispatcher(
            name="test_dispatcher_6021",
            map_log_level_func=lambda msg: 0,  # not used in test since we log directly to loggers,
            log_settings=LogSettings(verbosity=3, log_file_path=None, no_color=True),
        )
        caplog.clear()  # Clear any logs from LogDispatcher initialization
        with log_dispatcher.prefix_stderr_name(prefix="context_prefix_6021"):
            log_dispatcher.logger_stderr.info("test_6021")

        log_dispatcher.logger_stderr.info("test_6022")

        assert len(caplog.records) == 2
        assert caplog.records[0].name == "context_prefix_6021/test_dispatcher_6021.err"
        assert caplog.records[1].name == "test_dispatcher_6021.err"


class Test_DetermineLogLevel:
    @pytest.mark.parametrize(
        ["verbosity", "log_level"],
        [
            (-1000, logging.CRITICAL),
            (-2, logging.CRITICAL),
            (-1, logging.CRITICAL),
            (0, logging.WARNING),
            (1, logging.INFO),
            (2, logging.INFO),
            (3, logging.DEBUG),
            (4, logging.DEBUG),
            (1000, logging.DEBUG),
        ],
    )
    def test_correct_mapping_of_defined_log_levels(self, verbosity, log_level):
        assert LogDispatcher._determine_log_level(verbosity=verbosity) == log_level

    def test_error_on_not_equal_to_any_integer_but_comparable(self):
        with pytest.raises(ValueError) as exc_info:
            LogDispatcher._determine_log_level(verbosity=AlwaysCompareFalse())

        assert str(exc_info.value) == (
            "Somehow we ended up with a verbosity=AlwaysCompareFalseStub of "
            "type(verbosity)=<class 'cotainr.tests.tracing.stubs.AlwaysCompareFalse'> "
            "that does not compare well with integers."
        )



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

from pathlib import Path

from cotainr.cli import CotainrSubcommand
from cotainr.tracing import LogSettings


class StubDummyCLI:
    class DummySubcommand:
        def execute(self):
            print("Executed dummy subcommand.")

    def __init__(self, *, args=None):
        self.subcommand = self.DummySubcommand()
        print("Initialized DummyCLI.")


class StubInvalidSubcommand:
    pass


class StubLogSettingsSubcommand(CotainrSubcommand):
    def __init__(self, *, verbosity, log_file_path, no_color):
        self.log_settings = LogSettings(
            verbosity=verbosity, log_file_path=log_file_path, no_color=no_color
        )

    @classmethod
    def add_arguments(cls, *, parser):
        parser.add_argument("--verbosity", type=int)
        parser.add_argument("--log-file-path", type=Path, default=None)
        parser.add_argument("--no-color", action="store_true")

    def execute(self):
        pass


class StubValidSubcommand(CotainrSubcommand):
    def __init__(self, *, pos_arg, kw_arg=None):
        self.pos_arg = pos_arg
        self.kw_arg = kw_arg

    @classmethod
    def add_arguments(cls, *, parser):
        parser.add_argument("pos_arg")
        parser.add_argument("--kw-arg")

    def execute(self):
        print(
            f"Executed: '{self.__class__.__name__.lower()} "
            f"{self.pos_arg} --kw-arg={self.kw_arg}'",
            end="",
        )



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

Automatically create the list of release notes
"""

from pathlib import Path

release_notes_dir = Path(__file__).parent

print(f"Creating file {release_notes_dir.name}/release_list.rst")
with (release_notes_dir / "release_list.rst").open(mode="w") as f:
    for path in sorted(release_notes_dir.glob("*.md"), reverse=True):
        f.write(f".. include:: {path.name}\n")
        f.write("    :parser: myst_parser.sphinx_\n\n")



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import pathlib
from pathlib import Path

import pytest

import cotainr.util


@pytest.fixture
def patch_disable_stream_subprocess(monkeypatch):
    """
    Disable stream_subprocess(...).

    The `stream_subprocess` function is replaced by a mock that prints to
    stdout, logs to stderr (if a log_dispatcher is provided), and returns a
    message about the process that would have been run.
    """

    def mock_stream_subprocess(*, args, log_dispatcher, **kwargs):
        msg = f"PATCH: Streamed subprocess: {args=}, {kwargs=}"
        if log_dispatcher is not None:
            log_dispatcher.log_to_stderr(msg)
        print(msg)
        return msg

    monkeypatch.setattr(cotainr.util, "stream_subprocess", mock_stream_subprocess)


@pytest.fixture
def patch_empty_system(monkeypatch):
    """
    Fake an empty JSON file.

    Used to patch `cotainr.util.systems_file` as empty.
    """
    monkeypatch.setattr(pathlib.Path, "is_file", lambda _: True)
    monkeypatch.setattr(pathlib.Path, "read_text", lambda _: "{}")


@pytest.fixture
def patch_system_with_actual_file(monkeypatch):
    """
    Fake a JSON file with 2 entries.

    Used to patch `cotainr.util.systems_file` with two entries.
    """
    monkeypatch.setattr(pathlib.Path, "is_file", lambda _: True)
    monkeypatch.setattr(
        pathlib.Path,
        "read_text",
        lambda _: """{
            "some_system_6021": {"base-image": "some_base_image_6021"},
            "another_system_6021": {"base-image": "another_base_image_6021"}
        }""",
    )


@pytest.fixture
def patch_system_with_non_existing_file(monkeypatch):
    """
    Set `cotainr.util.systems_file` to an empty path.
    """
    monkeypatch.setattr(
        cotainr.util, "systems_file", Path("/some_non_existing_file_6021")
    )


@pytest.fixture
def patch_system_with_badly_formatted_file(monkeypatch):
    """
    Fake a badly formatted JSON file.

    Used to patch `cotainr.util.systems_file` as badly formatted.
    """
    monkeypatch.setattr(pathlib.Path, "is_file", lambda _: True)
    monkeypatch.setattr(
        pathlib.Path,
        "read_text",
        lambda _: '{"some_system_6021": {"some_variable_6021": "some_value_6021"}}',
    )



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

This module implements packing of software into the container.

Classes
-------
CondaInstall
    A Conda installation in a container sandbox.
"""

import logging
from pathlib import Path
import random
import re
import subprocess
import sys
import time
import urllib.error
import urllib.request

from . import tracing, util

logger = logging.getLogger(__name__)


class CondaInstall:
    """
    A Conda installation in a container sandbox.

    Bootstraps a Miniforge based Conda installation in a container sandbox. As
    part of the bootstrapping of Miniforge, the user must accept the `Miniforge
    license terms
    <https://github.com/conda-forge/miniforge/blob/main/LICENSE>`_.

    Parameters
    ----------
    sandbox : :class:`~cotainr.container.SingularitySandbox`
        The sandbox in which Conda should be installed.
    prefix : str
        The Conda prefix to use for the Conda install.
    license_accepted : bool, default=False
        The flag to indicate whether or not the user has already accepted the
        Miniforge license terms.
    log_settings : :class:`~cotainr.tracing.LogSettings`, optional
        The data used to setup the logging machinery (the default is None which
        implies that the logging machinery is not used).

    Attributes
    ----------
    sandbox : :class:`~cotainr.container.SingularitySandbox`
        The sandbox in which Conda is installed.
    prefix : str
        The Conda prefix used for the Conda install.
    license_accepted : bool
        Whether or not the Miniforge license terms have been accepted.
    log_dispatcher : :class:`~cotainr.tracing.LogDispatcher` or None.
        The log dispatcher used to process stdout/stderr message from
        Singularity commands that run in sandbox, if the logging machinery is
        used.

    Notes
    -----
    When adding a Conda environment, it is the responsibility of the user of
    cotainr to make sure they have the necessary rights to use the Conda
    channels/repositories and packages specified in the Conda environment, e.g.
    if `using the default Anaconda repositories
    <https://www.anaconda.com/blog/anaconda-commercial-edition-faq>`_.
    """

    def __init__(
        self,
        *,
        sandbox,
        prefix="/opt/cotainr/conda",
        license_accepted=False,
        log_settings=None,
    ):
        """Bootstrap a conda installation."""
        self.sandbox = sandbox
        self.prefix = prefix
        self.license_accepted = license_accepted
        if log_settings is not None:
            self._verbosity = log_settings.verbosity
            self.log_dispatcher = tracing.LogDispatcher(
                name=__class__.__name__,
                map_log_level_func=self._map_log_level,
                filters=self._logging_filters,
                log_settings=log_settings,
            )
        else:
            self._verbosity = 0
            self.log_dispatcher = None

        # Download Miniforge installer
        conda_installer_path = (
            Path(self.sandbox.sandbox_dir).resolve() / "conda_installer.sh"
        )
        self._download_miniforge_installer(installer_path=conda_installer_path)

        # Make sure the user has accepted the Miniforge installer license
        if not license_accepted:
            self._display_miniforge_license_for_acceptance(
                installer_path=conda_installer_path
            )
        else:
            self._display_message(
                msg=(
                    "You have accepted the Miniforge installer license via the command "
                    "line option '--accept-licenses'."
                ),
                log_level=logging.WARNING,
            )

        # Bootstrap Conda environment in container
        self._bootstrap_conda(installer_path=conda_installer_path)

        # Remove unneeded files
        conda_installer_path.unlink()
        self.cleanup_unused_files()

    def add_environment(self, *, path, name):
        """
        Add an exported Conda environment to the Conda install.

        Equivalent to calling "conda env create -f `path` -n `name`".

        Parameters
        ----------
        path : :class:`os.PathLike`
            The path to the exported env.yml file describing the Conda
            environment to install.
        name : str
            The name to use for the installed Conda environment.
        """
        self._run_command_in_sandbox(
            cmd=f"conda env create -f {path} -n {name}" + self._conda_verbosity_arg
        )

    def cleanup_unused_files(self):
        """
        Remove all unused Conda files.

        Equivalent to calling "conda clean -a".
        """
        self._run_command_in_sandbox(
            cmd="conda clean -y -a" + self._conda_verbosity_arg
        )

    def _bootstrap_conda(self, *, installer_path):
        """
        Install Conda and at its source script to the sandbox env.

        Parameters
        ----------
        installer_path : pathlib.Path
            The path of the Conda installer to run to bootstrap Conda.
        """
        # Run Conda installer
        self._run_command_in_sandbox(
            cmd=f"bash {installer_path.name} -b -s -p {self.prefix}"
        )

        # Add Conda to container sandbox env
        self.sandbox.add_to_env(
            shell_script=f"source {self.prefix + '/etc/profile.d/conda.sh'}"
        )

        # Check that we correctly use the newly installed Conda from now on
        self._check_conda_bootstrap_integrity()

        # Update the installed Conda package manager to the latest version
        self._run_command_in_sandbox(
            cmd=(
                "conda update -y -n base -c conda-forge conda"
                + self._conda_verbosity_arg
            )
        )

    def _check_conda_bootstrap_integrity(self):
        """Raise RuntimeError if multiple interfering Conda installs are found."""
        source_check_process = self._run_command_in_sandbox(cmd="conda info --base")
        if source_check_process.stdout.strip() != f"{self.prefix}":
            raise RuntimeError(
                "Multiple Conda installs interfere. "
                "We risk destroying the Conda install in "
                f"{source_check_process.stdout.strip()}. Aborting!"
            )

    def _display_message(self, *, msg, log_level=None):
        """
        Display a message to the user.

        Displays the message using the `log_dispatcher` if `log_level` is not
        `None` and a `log_dispatcher` is defined for the `CondaInstall`.
        Otherwise prints the message on stdout. When the `log_dispatcher` is
        used, messages with `log_levels` of WARNING or above are sent to stderr
        whereas massages with with `log_level` below WARNING are sent to
        stdout.

        Parameters
        ----------
        msg : str
            The message to display to the user.
        log_level : int, optional
            The logging level to use for the message, e.g. `logging.INFO` or
            `logging.WARNING`.
        """
        if self.log_dispatcher is None or log_level is None:
            print(msg)
        else:
            if log_level >= logging.WARNING:
                self.log_dispatcher.logger_stderr.log(level=log_level, msg=msg)
            else:
                self.log_dispatcher.logger_stdout.log(level=log_level, msg=msg)

    def _display_miniforge_license_for_acceptance(self, *, installer_path):
        """
        Extract and display Miniforge installer license for acceptance.

        Runs the Miniforge bootstrap installer to extract the license, displays
        it to the user, and prompts for acceptance of the license terms. Exits
        if the license terms are not accepted.

        Parameters
        ----------
        installer_path : pathlib.Path
            The path of the Miniforge installer to run to bootstrap Conda.

        Raises
        ------
        RuntimeError
            If unable to extract a license from the Miniforge installer.

        Notes
        -----
        This assumes that the Miniforge installer, as it is run, prompts the
        user for pressing ENTER to display the license, then displays the
        license and prompts the user to answer "yes" to accept the license
        terms.

        We try to "forward" this flow to the user by extracting the text shown
        when running the installer and pressing ENTER. We then prompt for a
        "yes" to the license terms.
        """
        with subprocess.Popen(
            ["bash", f"{installer_path.name}"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            text=True,
        ) as process:
            license_text, _ = process.communicate(
                # "press" ENTER to display the license and capture it
                "\n"
            )
            process.kill()  # We only use this process to extract the license

        util._flush_stdin_buffer()
        if license_text:
            license_text = license_text.replace(
                # remove prompt for pressing enter (as we have already done this...)
                "Please, press ENTER to continue\n>>> ",
                "\n",
            )
            # Remove "[yes|no]"" and ">>>" from license text as answer_is_yes
            # adds them as part of the input handling
            license_text = license_text.replace(" [yes|no]\n>>> ", "")
            logger.debug(f"The Miniforge displayed license is: {license_text}")
            # prompt user for acceptance of license terms
            if not util.answer_is_yes(license_text):
                self._display_message(
                    msg="You have not accepted the Miniforge installer license. Aborting!",
                    log_level=logging.CRITICAL,
                )
                sys.exit(0)

            self.license_accepted = True
            self._display_message(
                msg="You have accepted the Miniforge installer license.",
                log_level=logging.INFO,
            )
        else:
            raise RuntimeError(
                "No license seems to be displayed by the Miniforge installer."
            )

    @staticmethod
    def _get_install_script(architecture):
        """
        Determine the Miniforge installer to be downloaded based on system architecture.

        Always downloads a Linux version as the container is expected to always be Linux.

        Parameters
        ----------
        architecture : str
            The container architecture as returned by "uname -m".

        Raises
        ------
        ValueError
            If the container sandbox architecture is not supported.
        """
        if architecture in ("arm64", "aarch64"):
            install_script = "Miniforge3-Linux-aarch64.sh"
        elif architecture == "x86_64":
            install_script = "Miniforge3-Linux-x86_64.sh"
        else:
            raise ValueError(
                "Cotainr's CondaInstall only supports x86_64 and arm64/aarch64. "
                f'Cotainr got "{architecture=}" for your container'
            )

        return install_script

    def _download_miniforge_installer(self, *, installer_path):
        """
        Download the Miniforge installer to `installer_path`.

        Parameters
        ----------
        installer_path : pathlib.Path
            The path to download the conda installer to.

        Raises
        ------
        RuntimeError
            If the container sandbox architecture is unknown.
        urllib.error.URLError
            If three attempts at downloading the installer all fail.
        """
        architecture = self.sandbox.architecture
        if architecture is None:
            raise RuntimeError(
                f"Cotainr's CondaInstall got '{architecture=}' "
                "which indicates that it is not running in a container sandbox context."
            )

        install_script = CondaInstall._get_install_script(architecture)
        miniforge_installer_url = (
            "https://github.com/conda-forge/miniforge/releases/latest/download/"
            + install_script
        )

        # Make up to 3 attempts at downloading the installer
        for retry in range(3):
            try:
                with urllib.request.urlopen(miniforge_installer_url) as url:  # nosec B310
                    installer_path.write_bytes(url.read())

                break

            except urllib.error.URLError as e:
                url_error = e

                # Exponential back-off
                time.sleep(2**retry + random.uniform(0.001, 1))  # nosec B311

        else:
            raise url_error

    def _run_command_in_sandbox(self, *, cmd):
        """
        Wrap the sandbox command runner to use class specific log_dispatcher.

        Wraps calls to `self.sandbox.run_command_in_container` to use
        self.log_dispatcher for log handling instead of the sandbox's log
        dispatcher.

        Parameters
        ----------
        cmd : str
            The command to run in the container sandbox.

        Returns
        -------
        process : :class:`subprocess.CompletedProcess`
            Information about the process that ran in the container sandbox.
        """
        return self.sandbox.run_command_in_container(
            cmd=cmd, custom_log_dispatcher=self.log_dispatcher
        )

    @property
    def _conda_verbosity_arg(self):
        """
        Get a verbosity level for Conda commands.

        A mapping of the internal cotainr verbosity level to `Conda verbosity
        flags
        <https://docs.conda.io/projects/conda/en/latest/commands/create.html#Output,%20Prompt,%20and%20Flow%20Control%20Options>`_.

        Returns
        -------
        verbosity_arg : str
            The verbosity arg ("-q", "-v", "-vv", etc.) to add to the Conda
            command.
        """
        if self._verbosity <= 0:
            return " -q"
        elif self._verbosity == 2:
            # Conda INFO
            return " -v"
        elif self._verbosity == 3 or self._verbosity == 4:
            # Conda DEBUG
            return " -vv"
        elif self._verbosity >= 5:
            # Conda TRACE
            return " -vvv"
        else:
            return ""

    @property
    def _logging_filters(self):
        """
        Create logging filters for messages from conda commands.

        Returns
        -------
        logging_filters : list of :class:`logging.Filter`.
            The list of filters to use with the logging machinery when handling
            messages from conda commands.
        """

        class StripANSIEscapeCodes(logging.Filter):
            """
            In-place strip all ANSI escape codes.

            Regex from https://stackoverflow.com/a/14693789
            """

            ansi_escape_re = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

            def filter(self, record):
                record.msg = self.ansi_escape_re.sub("", record.msg)
                return True

        class NoEmptyLinesFilter(logging.Filter):
            """Remove any empty lines."""

            def filter(self, record):
                return record.msg.strip() != ""

        class OnlyFinalProgressbarFilter(logging.Filter):
            """
            Only include final 100% complete line when download progress bars are shown.

            Assume a progress bar line like
            [some text]|[some text]|[progress bar characters]| [percentage complete]% [ansi escape codes]
            """

            progress_bar_re = re.compile(
                r"^(.+?)\|(.+?)\|[ \#0-9]+?\|[ ]{1,3}[0-9]{1,2}\%"
            )

            def filter(self, record):
                return not self.progress_bar_re.match(record.msg)

        logging_filters = [
            # The order matters as filters are applied in order. ANSI escape
            # codes and partial progress bars must be removed before filtering
            # empty lines.
            StripANSIEscapeCodes(),
            OnlyFinalProgressbarFilter(),
            NoEmptyLinesFilter(),
        ]

        return logging_filters

    @staticmethod
    def _map_log_level(msg):
        """
        Attempt to infer log level for a message.

        Parameters
        ----------
        msg : str
            The message to infer log level for.

        Returns
        -------
        log_level : int
            One of the standard log levels (DEBUG, INFO, WARNING, ERROR, or
            CRITICAL).
        """
        if (
            msg.startswith("DEBUG")
            or msg.startswith("VERBOSE")
            or msg.startswith("TRACE")
        ):
            return logging.DEBUG
        elif msg.startswith("INFO"):
            return logging.INFO
        elif msg.startswith("WARNING"):
            return logging.WARNING
        elif msg.startswith("ERROR"):
            return logging.ERROR
        elif msg.startswith("CRITICAL"):
            return logging.CRITICAL
        else:
            # If no prefix on message, assume its INFO level
            return logging.INFO



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import io
import shutil
import time

import pytest

from cotainr.tracing import MessageSpinner

from .stubs import FixedNumberOfSpinsEvent


@pytest.fixture()
def safe_MessageSpinner(msg, stream):
    """
    Setup a MessageSpinner and make sure to stop it during teardown.

    The parameters `msg` and `stream` must be passed to the fixture using a
    pytest parameterization.
    """
    safe_MessageSpinner = MessageSpinner(msg=msg, stream=stream)
    yield safe_MessageSpinner
    safe_MessageSpinner.stop()


class TestStart:
    @pytest.mark.parametrize(["msg", "stream"], [("test 6021", io.StringIO())])
    def test_correct_start(self, msg, stream, safe_MessageSpinner):
        # The `msg` and `stream` parameters are passed automagically to the
        # `safe_MessageSpinner` fixture.
        assert not safe_MessageSpinner._running
        assert not safe_MessageSpinner._spinner_thread.is_alive()
        safe_MessageSpinner.start()
        assert safe_MessageSpinner._running
        assert safe_MessageSpinner._spinner_thread.is_alive()

    @pytest.mark.parametrize(["msg", "stream"], [("test 6021", io.StringIO())])
    def test_spinning_started(self, msg, stream, safe_MessageSpinner):
        # The `msg` and `stream` parameters are passed automagically to the
        # `safe_MessageSpinner` fixture.
        safe_MessageSpinner.start()
        for _ in range(5):
            # wait for something to be written to the stream
            stream_msg = stream.getvalue()
            if stream_msg:
                break
            else:
                time.sleep(0.2)
        else:
            raise RuntimeError("No message seems to be spinning")

        assert msg in stream_msg


class TestStop:
    @pytest.mark.parametrize(["msg", "stream"], [("test 6021", io.StringIO())])
    def test_stop_before_having_started(self, msg, stream, safe_MessageSpinner):
        # The `msg` and `stream` parameters are passed automagically to the
        # `safe_MessageSpinner` fixture.
        assert not safe_MessageSpinner._running
        assert not safe_MessageSpinner._spinner_thread.is_alive()
        safe_MessageSpinner.stop()
        assert not safe_MessageSpinner._running
        assert not safe_MessageSpinner._spinner_thread.is_alive()

    @pytest.mark.parametrize(["msg", "stream"], [("test 6021", io.StringIO())])
    def test_stop_when_running(self, msg, stream, safe_MessageSpinner):
        # The `msg` and `stream` parameters are passed automagically to the
        # `safe_MessageSpinner` fixture.
        assert not safe_MessageSpinner._running
        assert not safe_MessageSpinner._spinner_thread.is_alive()
        safe_MessageSpinner.start()
        assert safe_MessageSpinner._running
        assert safe_MessageSpinner._spinner_thread.is_alive()
        safe_MessageSpinner.stop()
        assert not safe_MessageSpinner._running
        assert not safe_MessageSpinner._spinner_thread.is_alive()


class Test_SpinMsg:
    @pytest.mark.parametrize(
        ["msg", "stream", "no_ansi_msg"],
        [
            (
                # non SGR escape codes
                "\033[Asome_long\x1b[A_message_w\x1b[2;Eith_a_\x1b[5;3;Hlot_of_"
                "AN\x1b[KSI_codes_\x1b[B6021",
                io.StringIO(),
                "some_long_message_with_a_lot_of_ANSI_codes_6021",
            ),
            (
                # "message_with_SGRs"
                "\x1b[38;5;8mmess\x1b[38;5;8mage_with\x1b[38;5;1m_\x1b[38;5;160mSGRs"
                "\x1b[0m",
                io.StringIO(),
                "\x1b[38;5;8mmess\x1b[38;5;8mage_with\x1b[38;5;1m_\x1b[38;5;160mSGRs"
                "\x1b[0m",
            ),
        ],
    )
    def test_ansi_escape_codes_removal(
        self, msg, stream, no_ansi_msg, safe_MessageSpinner, monkeypatch
    ):
        # The `msg` and `stream` parameters are passed automagically to the
        # `safe_MessageSpinner` fixture.
        monkeypatch.setattr(
            safe_MessageSpinner, "_stop_signal", FixedNumberOfSpinsEvent(spins=0)
        )
        safe_MessageSpinner._spin_msg()

        stream_msg = stream.getvalue().rstrip("\n")
        # This implements stream_msg.removeprefix() which is found in Python >= 3.9
        if stream_msg.startswith(safe_MessageSpinner._clear_line_code):
            stream_msg = stream_msg[len(safe_MessageSpinner._clear_line_code) :]

        assert stream_msg == no_ansi_msg

    @pytest.mark.parametrize(
        ["msg", "stream", "no_newline_msg"],
        [
            ("test 6021\n", io.StringIO(), "test 6021"),
            ("test 6021\n\n", io.StringIO(), "test 6021"),
            ("\ntest\n 60\n21", io.StringIO(), "\ntest\n 60\n21"),
            ("\ntest\n 60\n21\n", io.StringIO(), "\ntest\n 60\n21"),
            (
                "\x1b[38;5;8mtest 6021\x1b[0m\n",
                io.StringIO(),
                "\x1b[38;5;8mtest 6021\x1b[0m",
            ),
            (
                "\x1b[38;5;8mtest 6021\n\x1b[0m",
                io.StringIO(),
                "\x1b[38;5;8mtest 6021\x1b[0m",
            ),
            (
                "\x1b[38;5;8mtest 6021\n\n\x1b[0m",
                io.StringIO(),
                "\x1b[38;5;8mtest 6021\x1b[0m",
            ),
            (
                "\n\x1b[38;5;8m\ntest 60\n21\n\n\x1b[0m",
                io.StringIO(),
                "\n\x1b[38;5;8m\ntest 60\n21\x1b[0m",
            ),
        ],
    )
    def test_newline_at_end_removal(
        self, msg, stream, no_newline_msg, safe_MessageSpinner, monkeypatch
    ):
        # The `msg` and `stream` parameters are passed automagically to the
        # `safe_MessageSpinner` fixture.
        monkeypatch.setattr(
            safe_MessageSpinner, "_stop_signal", FixedNumberOfSpinsEvent(spins=0)
        )
        safe_MessageSpinner._spin_msg()

        stream_msg = stream.getvalue().rstrip("\n")
        # This implements stream_msg.removeprefix() which is found in Python >= 3.9
        if stream_msg.startswith(safe_MessageSpinner._clear_line_code):
            stream_msg = stream_msg[len(safe_MessageSpinner._clear_line_code) :]

        assert stream_msg == no_newline_msg

    @pytest.mark.parametrize(["msg", "stream"], [("test 6021", io.StringIO())])
    def test_spinner_cycle(self, msg, stream, safe_MessageSpinner, monkeypatch):
        # The `msg` and `stream` parameters are passed automagically to the
        # `safe_MessageSpinner` fixture.
        monkeypatch.setattr(
            safe_MessageSpinner,
            "_stop_signal",
            FixedNumberOfSpinsEvent(spins=16),  # Two full cycles
        )
        expected_spin_output = (
            "\r\x1b[2K⣾ test 6021..."
            "\r\x1b[2K⣷ test 6021..."
            "\r\x1b[2K⣯ test 6021..."
            "\r\x1b[2K⣟ test 6021..."
            "\r\x1b[2K⡿ test 6021..."
            "\r\x1b[2K⢿ test 6021..."
            "\r\x1b[2K⣻ test 6021..."
            "\r\x1b[2K⣽ test 6021..."
        ) * 2 + "\r\x1b[2Ktest 6021\n"
        safe_MessageSpinner._spin_msg()
        stream_msg = stream.getvalue()
        assert stream_msg == expected_spin_output

    @pytest.mark.parametrize(
        ["msg", "stream"],
        [
            (
                "some_message_longer_than_terminal_width"
                + "!" * shutil.get_terminal_size()[0],
                io.StringIO(),
            )
        ],
    )
    def test_one_line_msg(self, msg, stream, safe_MessageSpinner, monkeypatch):
        # The `msg` and `stream` parameters are passed automagically to the
        # `safe_MessageSpinner` fixture.
        monkeypatch.setattr(
            safe_MessageSpinner, "_stop_signal", FixedNumberOfSpinsEvent(spins=1)
        )
        exclamation_marks = shutil.get_terminal_size()[0] - len(
            "⣾ some_message_longer_than_terminal_width..."
        )
        expected_one_line_msg = (
            f"⣾ some_message_longer_than_terminal_width{'!' * exclamation_marks}..."
        )
        safe_MessageSpinner._spin_msg()
        one_line_stream_msg = stream.getvalue().split("\r\x1b[2K")[1]
        assert one_line_stream_msg == expected_one_line_msg

    @pytest.mark.parametrize(
        ["msg", "stream"],
        [
            (
                "some_message_longer_than_terminal_width"
                + "!" * shutil.get_terminal_size()[0],
                io.StringIO(),
            )
        ],
    )
    def test_print_final_full_message(
        self, msg, stream, safe_MessageSpinner, monkeypatch
    ):
        # The `msg` and `stream` parameters are passed automagically to the
        # `safe_MessageSpinner` fixture.
        monkeypatch.setattr(
            safe_MessageSpinner, "_stop_signal", FixedNumberOfSpinsEvent(spins=1)
        )
        safe_MessageSpinner._spin_msg()
        final_stream_msg = stream.getvalue().split("\r\x1b[2K")[-1]
        assert final_stream_msg == msg + "\n"



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import os
from pathlib import Path
import sys
import time

sys.path.insert(0, f"{Path('..').resolve()}")

import cotainr

project = cotainr.__name__
author = "DeiC"
copyright = f"2022-{time.strftime('%Y')}, {author}"
version = cotainr.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration
# https://numpydoc.readthedocs.io/en/latest/install.html#configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.duration",
    "sphinx.ext.intersphinx",
    "sphinx_design",
    "numpydoc",
    "myst_parser",
]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
templates_path = ["_templates"]
intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}
numpydoc_class_members_toctree = False

# -- Options for linkcheck ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-the-linkcheck-builder
linkcheck_retries = 3  # Retry 3 times before considering a link broken
linkcheck_workers = 1  # Limit the number of workers to avoid hitting rate limits
linkcheck_ignore = [
    # It is easy to hit the rate limit of stackoverflow
    # Let's assume that stackoverflows does not move its supposedly permanent links
    "https://stackoverflow.com",
    # Its also easy to hit the rate limit of github
    # We do have quite a lot of links to our own github repo
    # For now, let's avoid checking these
    "https://github.com/DeiC-HPC/cotainr",
    # Also ignore everything on the gnu domain via regex.
    r"https://.*\.gnu\.org/.*",
    # Ignore the links to specific versions of the cotainr docs.
    # This is required because otherwise the link checking fails on releasing a new version of cotainr,
    # as the docs are not yet available when doing the pull request.
    r"https:\/\/cotainr\.readthedocs\.org\/en\/([0-9]+(\.[0-9]+)*)\/",
]
linkcheck_anchors_ignore_for_url = [
    # Ignore GitHub issue comment anchors that apparently fails
    "https://github.com/apptainer/singularity/issues/5941"
]
if "GITHUB_TOKEN" in os.environ:
    # Authenticate with GitHub token in GitHub Actions to avoid hitting rate limits
    # It is not entirely clear if this has any effect outside of api.github.com, though...
    linkcheck_request_headers = {
        "https://github.com": {"Authorization": f"token {os.environ['GITHUB_TOKEN']}"}
    }

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

rtd_version = os.environ.get("READTHEDOCS_VERSION")
if not rtd_version or rtd_version.isdigit():
    # Either not on readthedocs or it is a pull request number
    switcher_version = "latest"
    switcher_json = "_static/switcher.json"
else:
    switcher_version = rtd_version
    switcher_json = "https://cotainr.readthedocs.io/en/latest/_static/switcher.json"

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_css_files = ["css/cotainr.css"]
html_title = project
html_context = {
    "github_user": "DeiC-HPC",
    "github_repo": project,
    "github_version": "main",
    "doc_path": "doc",
}
html_theme_options = {
    "use_edit_page_button": True,
    "navbar_start": ["navbar-logo", "version-switcher"],
    "footer_start": ["cotainr_footer"],
    "switcher": {
        "json_url": switcher_json,
        "version_match": switcher_version,
    },
    "icon_links": [
        {
            "name": "GitHub",
            "url": f"https://github.com/{html_context['github_user']}/{html_context['github_repo']}",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        },
        {
            "name": "PyPI",
            "url": f"https://pypi.org/project/{project}/",
            "icon": "fa-brands fa-python",
            "type": "fontawesome",
        },
    ],
    # We only set "navigation_with_keys = False" to explicitly remove the
    # warning introduced in pydata-sphinx-theme v0.14.2:
    # "WARNING: The default value for `navigation_with_keys` will change to
    # `False` in the next release. If you wish to preserve the old behavior for
    # your site, set `navigation_with_keys=True` in the `html_theme_options`
    # dict in your `conf.py` file.Be aware that `navigation_with_keys = True`
    # has negative accessibility
    # implications:https://github.com/pydata/pydata-sphinx-theme/issues/1492"
    "navigation_with_keys": False,
}

# -- Autodoc configuration ----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration


def add_api_headline_to_module_docs(app, what, name, obj, options, lines):
    """Add an 'API reference' headline before showing API content in auto modules."""
    if what == "module":
        lines.append("\n")
        lines.append("API reference")
        lines.append("-------------")


def setup(app):  # noqa: D103
    app.connect("autodoc-process-docstring", add_api_headline_to_module_docs)



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

from abc import ABC


class StubLicensePopen(ABC):
    license_text: str

    def __init__(self, args, stdin=None, stdout=None, text=None):  # noqa: B027
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):  # noqa: B027
        pass

    def communicate(self, input=None):
        print(f"{self.__class__.__name__} received: {input=}.")
        return (self.license_text, "")

    def kill(self):
        print(f"{self.__class__.__name__} killed.")


class StubEmptyLicensePopen(StubLicensePopen):
    license_text = ""


class StubShowLicensePopen(StubLicensePopen):
    license_text = (
        "STUB:\nPlease, press ENTER to continue\n>>> \nThis is the license terms...\n"
    )



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import pytest

from cotainr.cli import _extract_help_from_docstring


class TestExtractHelpFromDocstring:
    def test_arg_not_found(self):
        docstring = """
            Summary

            Parameters
            ----------
            foo : str
                Foo

            Returns
            -------
            bar : str
                Bar
            """
        with pytest.raises(
            KeyError, match="The docstring does not include arg='some_arg'"
        ):
            _extract_help_from_docstring(arg="some_arg", docstring=docstring)

    def test_description_first_letter_lower_case(self):
        docstring = """
        Parameters
        ----------
        some_arg : int
            The INTEGER
        """
        help_msg = _extract_help_from_docstring(arg="some_arg", docstring=docstring)
        assert help_msg == "the INTEGER"

    def test_description_period_strip(self):
        docstring = """
        Parameters
        ----------
        some_arg : float
            A float.
        """
        help_msg = _extract_help_from_docstring(arg="some_arg", docstring=docstring)
        assert help_msg == "a float"

    def test_description_whitespace_strip(self):
        docstring = """
        Parameters
        ----------
        some_arg : str
            Much leading and trailing space
        """
        # Doctor the docstring a little here, so editors etc. which eagerly
        # strip whitespace don't mess up the test.
        docstring = docstring.replace("space", "space" + " " * 10)
        help_msg = _extract_help_from_docstring(arg="some_arg", docstring=docstring)
        assert help_msg == "much leading and trailing space"

    def test_multi_line_arg_description(self):
        docstring = """
        Parameters
        ----------
        some_arg : list
            Some description of this list
            on multiple lines
        """
        help_msg = _extract_help_from_docstring(arg="some_arg", docstring=docstring)
        assert help_msg == "some description of this list on multiple lines"

    def test_no_param_section(self):
        docstring = """No args here..."""
        with pytest.raises(
            KeyError, match="The docstring does not include arg='some_arg'"
        ):
            _extract_help_from_docstring(arg="some_arg", docstring=docstring)

    def test_single_line_arg_description(self):
        docstring = """
        Parameters
        ----------
        some_arg : tuple
            Some description of this tuple on one line.

        Returns
        -------
        something : None
            Not really anything...
        """
        help_msg = _extract_help_from_docstring(arg="some_arg", docstring=docstring)
        assert help_msg == "some description of this tuple on one line"



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import pytest


@pytest.fixture
def data_log_dispatcher_critical_color_log_messages(data_log_level_names_mapping):
    """
    Log messages for standard log levels and their expected output to
    stdout/stderr for the LogDispatcher logging machinery at CRITICAL level.
    """
    # Choose a logger name
    logger_name = "LD_6021"

    # Messages to log at the different standard log levels
    log_level_msgs = {
        level: f"Log {level_name} 6021"
        for level, level_name in data_log_level_names_mapping.items()
    }

    # Expected stdout messages
    stdout_msgs = ["LD_6021.out:-: \x1b[38;5;160mLog CRITICAL 6021\x1b[0m"]

    # Expected stderr messages
    stderr_msgs = ["LD_6021.err:-: \x1b[38;5;160mLog CRITICAL 6021\x1b[0m"]

    return logger_name, log_level_msgs, stdout_msgs, stderr_msgs


@pytest.fixture
def data_log_dispatcher_debug_color_log_messages(data_log_level_names_mapping):
    """
    Log messages for standard log levels and their expected output to
    stdout/stderr for the LogDispatcher logging machinery at DEBUG level.
    """
    # Choose a logger name
    logger_name = "LD_6021"

    # Messages to log at the different standard log levels
    log_level_msgs = {
        level: f"Log {level_name} 6021"
        for level, level_name in data_log_level_names_mapping.items()
    }

    # Expected ending of stdout messages
    # (debug messages are prepended with a time stamp)
    stdout_msgs = [
        "LD_6021.out:-:CRITICAL: \x1b[38;5;160mLog CRITICAL 6021\x1b[0m",
        "LD_6021.out:-:ERROR: \x1b[38;5;1mLog ERROR 6021\x1b[0m",
        "LD_6021.out:-:WARNING: \x1b[38;5;3mLog WARNING 6021\x1b[0m",
        "LD_6021.out:-:INFO: Log INFO 6021",
        "LD_6021.out:-:DEBUG: \x1b[38;5;8mLog DEBUG 6021\x1b[0m",
    ]

    # Expected ending of stderr messages
    # (debug messages are prepended with a time stamp)
    stderr_msgs = [
        "LD_6021.err:-:CRITICAL: \x1b[38;5;160mLog CRITICAL 6021\x1b[0m",
        "LD_6021.err:-:ERROR: \x1b[38;5;1mLog ERROR 6021\x1b[0m",
        "LD_6021.err:-:WARNING: \x1b[38;5;3mLog WARNING 6021\x1b[0m",
        "LD_6021.err:-:INFO: Log INFO 6021",
        "LD_6021.err:-:DEBUG: \x1b[38;5;8mLog DEBUG 6021\x1b[0m",
    ]

    return logger_name, log_level_msgs, stdout_msgs, stderr_msgs


@pytest.fixture
def data_log_dispatcher_info_color_log_messages(data_log_level_names_mapping):
    """
    Log messages for standard log levels and their expected output to
    stdout/stderr for the LogDispatcher logging machinery at INFO level.
    """
    # Choose a logger name
    logger_name = "LD_6021"

    # Messages to log at the different standard log levels
    log_level_msgs = {
        level: f"Log {level_name} 6021"
        for level, level_name in data_log_level_names_mapping.items()
    }

    # Expected stdout messages
    stdout_msgs = [
        "LD_6021.out:-: \x1b[38;5;160mLog CRITICAL 6021\x1b[0m",
        "LD_6021.out:-: \x1b[38;5;1mLog ERROR 6021\x1b[0m",
        "LD_6021.out:-: \x1b[38;5;3mLog WARNING 6021\x1b[0m",
        "LD_6021.out:-: Log INFO 6021",
    ]

    # Expected stderr messages
    stderr_msgs = [
        "LD_6021.err:-: \x1b[38;5;160mLog CRITICAL 6021\x1b[0m",
        "LD_6021.err:-: \x1b[38;5;1mLog ERROR 6021\x1b[0m",
        "LD_6021.err:-: \x1b[38;5;3mLog WARNING 6021\x1b[0m",
        "LD_6021.err:-: Log INFO 6021",
    ]

    return logger_name, log_level_msgs, stdout_msgs, stderr_msgs


@pytest.fixture
def data_log_dispatcher_info_no_color_log_messages(data_log_level_names_mapping):
    """
    Log messages for standard log levels and their expected output to
    stdout/stderr for the LogDispatcher logging machinery at INFO level -
    without message coloring.
    """
    # Choose a logger name
    logger_name = "LD_6021"

    # Messages to log at the different standard log levels
    log_level_msgs = {
        level: f"Log {level_name} 6021"
        for level, level_name in data_log_level_names_mapping.items()
    }

    # Expected stdout messages
    stdout_msgs = [
        "LD_6021.out:-: Log CRITICAL 6021",
        "LD_6021.out:-: Log ERROR 6021",
        "LD_6021.out:-: Log WARNING 6021",
        "LD_6021.out:-: Log INFO 6021",
    ]

    # Expected stderr messages
    stderr_msgs = [
        "LD_6021.err:-: Log CRITICAL 6021",
        "LD_6021.err:-: Log ERROR 6021",
        "LD_6021.err:-: Log WARNING 6021",
        "LD_6021.err:-: Log INFO 6021",
    ]

    return logger_name, log_level_msgs, stdout_msgs, stderr_msgs


@pytest.fixture
def data_log_dispatcher_warning_color_log_messages(data_log_level_names_mapping):
    """
    Log messages for standard log levels and their expected output to
    stdout/stderr for the LogDispatcher logging machinery at WARNING level.
    """
    # Choose a logger name
    logger_name = "LD_6021"

    # Messages to log at the different standard log levels
    log_level_msgs = {
        level: f"Log {level_name} 6021"
        for level, level_name in data_log_level_names_mapping.items()
    }

    # Expected stdout messages
    stdout_msgs = [
        "LD_6021.out:-: \x1b[38;5;160mLog CRITICAL 6021\x1b[0m",
        "LD_6021.out:-: \x1b[38;5;1mLog ERROR 6021\x1b[0m",
        "LD_6021.out:-: \x1b[38;5;3mLog WARNING 6021\x1b[0m",
    ]

    # Expected stderr messages
    stderr_msgs = [
        "LD_6021.err:-: \x1b[38;5;160mLog CRITICAL 6021\x1b[0m",
        "LD_6021.err:-: \x1b[38;5;1mLog ERROR 6021\x1b[0m",
        "LD_6021.err:-: \x1b[38;5;3mLog WARNING 6021\x1b[0m",
    ]

    return logger_name, log_level_msgs, stdout_msgs, stderr_msgs



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk Licensed under the European Union Public License (EUPL)
1.2 - see the LICENSE file for details.

This script is used to prepare the release notes and version switcher for the
documentation for a new release of cotainr. It automatically generates the
version switcher for the documentation and creates a skeleton for the release
notes. The script should be run before making a tag for the latest version.
More information about releasing can be found here:
https://cotainr.readthedocs.io/en/latest/development/releasing.html

"""

import argparse
import datetime
import json
import locale
from pathlib import Path
import re
import subprocess
import sys
from typing import Optional

sys.path.insert(0, f"{(Path(__file__) / '../..').resolve()}")

import cotainr


def _check_release_version_format(*, release_version: str, raise_error: bool = True):
    """
    Check if the release version is in the correct YYYY.MM.MICRO format.

    For releases before April 2025, the version YYYY.0M.MICRO is also accepted
    as we had incorrectly used this format up until then.

    Parameters
    ----------
    release_version : str
        The version for the new release.
    raise_error : bool
        If True, raise a ValueError if the version is not in the correct
        format.

    Returns
    -------
    acceptable_format : bool
        True if the version is in the correct format, False otherwise.

    Raises
    ------
    ValueError
        If the release version is not in the correct format and raise_error is
        True.
    """
    correct_YYYY_MM_MICRO_version_format_re = (
        cotainr._version._get_cotainr_calver_tag_pattern()
    )
    wrong_YYYY_0M_MICRO_version_format_re = r"^20[0-9]{2}\.(0[1-9])\.(0|[1-9][0-9]*)$"

    if re.match(wrong_YYYY_0M_MICRO_version_format_re, release_version):
        # If the version is in the wrong YYYY.0M.MICRO format, accept it only
        # if the release date is before April 2025.
        year, month, _ = map(int, release_version.split("."))
        if datetime.date(year, month, 1) < datetime.date(2025, 4, 1):
            acceptable_format = True
        else:
            acceptable_format = False
    elif re.match(correct_YYYY_MM_MICRO_version_format_re, release_version):
        # If the version is not in the correct YYYY.MM.MICRO format, accept it
        acceptable_format = True
    else:
        acceptable_format = False

    if not acceptable_format and raise_error:
        raise ValueError(
            f"Release version {release_version} is not in the expected YYYY.MM.MICRO format."
        )

    return acceptable_format


def create_docs_switcher(*, formatted_release_version: str):
    """
    Create the version switcher for the documentation.

    Automatically create the version switcher configuration for the PyData
    Sphinx Theme. The switcher will contain the latest 4 releases. See
    https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/version-dropdown.html#add-a-json-file-to-define-your-switcher-s-versions
    for more technical details.

    Parameters
    ----------
    formatted_release_version : str
        The YYYY.MM.MICRO version tag for the new release.

    Returns
    -------
    switcher : str
        The pretty formatted JSON string representation of the version switcher
        configuration for the documentation.
    """
    _check_release_version_format(release_version=formatted_release_version)

    # Create base switcher with the latest and stable versions
    switcher = [
        {
            "name": "dev",
            "version": "latest",
            "url": "https://cotainr.readthedocs.io/en/latest/",
        },
        {
            "name": f"{formatted_release_version} (stable)",
            "version": "stable",
            "url": "https://cotainr.readthedocs.io/en/stable/",
        },
    ]

    # Append the latest 3 releases to the switcher based on git tags that match
    # the cotainr version format (or the wrong practice format)
    tags = [
        tag
        for tag in (
            subprocess.run(
                ["git", "--no-pager", "tag"], capture_output=True, text=True
            ).stdout.splitlines()
        )
        if _check_release_version_format(release_version=tag, raise_error=False)
    ]
    tags.reverse()
    for tag in tags[0:3]:
        switcher.append(
            {
                "name": tag,
                "version": tag,
                "url": f"https://cotainr.readthedocs.io/en/{tag}/",
            }
        )

    return json.dumps(switcher, indent="  ")


def create_release_notes(
    *, formatted_release_version: str, formatted_release_date: str
):
    """
    Create the release notes for the new version.

    Automatically create the skeleton for the release notes for the new
    version. Fills in the version number in the release note template.

    Parameters
    ----------
    formatted_release_version : str
        The YYYY.MM.MICRO version tag for the new release.
    formatted_release_date : str
        The release date formatted as "__Month__ __day__, __year__".

    Returns
    -------
    release_notes : str
        The release notes skeleton for the new version.
    """
    _check_release_version_format(release_version=formatted_release_version)

    release_notes = (
        (
            # Load release note template
            (Path(__file__) / "../release_notes/release_note.md.template")
            .resolve()
            .read_text()
        )
        .replace(
            # Insert the new version number as the title
            "__YYYY.MM.MICRO__",
            formatted_release_version,
        )
        .replace(
            # Insert the release date
            "__Month__ __day__, __year__",
            f"{formatted_release_date}",
        )
        .replace(
            # Insert version tag for RTD link
            "__version__tag__",
            formatted_release_version,
        )
    )

    return release_notes


def format_release_version_and_date(*, release_date: Optional[str] = None):
    """
    Get the release date and version from the command line arguments.

    Parameters
    ----------
    release_date : str or None
        The release date for the new version in ISO 8601 format, e.g. YYYY-MM-DD. If None,
        the current date is used.

    Returns
    -------
    formatted_release_ver : str
        The new version number in YYYY.MM.MICRO format.
    formatted_release_date : str
        The release date formatted as "__Month__ __day__, __year__".
    """
    if release_date is None:
        date_release = datetime.datetime.today().date()
    else:
        date_release = datetime.date.fromisoformat(release_date)

    # Extract the current development version number from the cotainr module.
    # Note: Micro version automatically increment.
    try:
        yyyy, mm, micro = cotainr.__version__.split(".")[:3]
    except ValueError as e:
        raise ValueError(
            f"Current release version {cotainr.__version__} is not in the expected format."
        ) from e

    _check_release_version_format(release_version=f"{yyyy}.{mm}.{micro}")

    # Determine the MICRO version number
    if date_release.year < int(yyyy) or (
        date_release.year == int(yyyy) and date_release.month < int(mm)
    ):
        raise ValueError(
            f"New release date {date_release} is before the current version {cotainr.__version__}."
        )
    elif date_release.year > int(yyyy) or date_release.month > int(mm):
        # New year or month since last release, set MICRO to 0
        micro_release_ver = 0
    else:
        # Same year/month as last release, set MICRO += 1
        micro_release_ver = int(micro)  # micro release is already updated by hatch-vcs.

    # Format the full new version number
    formatted_release_ver = (
        f"{date_release.year}.{date_release.month}.{micro_release_ver}"
    )

    # Format the release date for the release notes
    formatted_release_date = _format_release_notes_date(date_release)

    return formatted_release_ver, formatted_release_date


def _format_release_notes_date(date: datetime.date):
    """
    Format `date` as a "__Month__ __day__, __year__" string.

    Parameters
    ----------
    date : datetime.date
        The date to format.

    Returns
    -------
    formatted_date : str
        The formatted date string.
    """
    # Determine english orginal prefix
    if date.day in [1, 21, 31]:
        day_prefix = "st"
    elif date.day in [2, 22]:
        day_prefix = "nd"
    elif date.day in [3, 23]:
        day_prefix = "rd"
    else:
        day_prefix = "th"

    current_locale = locale.getlocale()
    if current_locale[0] not in ["C", "en_US", "en_GB"]:
        raise RuntimeError(
            f"Your locale is set to {current_locale}. "
            "Please set it to an English locale (C, en_US, or en_GB) "
            "via e.g. the LANG environment variable before running this "
            "script to ensure the release date in the release notes is "
            "formatted correctly."
        )

    formatted_date = date.strftime(f"%B %-d{day_prefix}, %Y")

    return formatted_date


if __name__ == "__main__":  # pragma: no cover
    # Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--release_date",
        type=str,
        help=(
            "The release date for the new version in ISO 8601 format, e.g. YYYY-MM-DD. "
            "If not provided, the current date is used."
        ),
    )
    args = parser.parse_args()

    # Format the release version number and release notes date
    formatted_release_version, formatted_release_date = format_release_version_and_date(
        release_date=args.release_date
    )

    # Create the new version switcher and write it to a static file
    version_switcher = create_docs_switcher(
        formatted_release_version=formatted_release_version
    )
    version_switcher_file = Path(__file__).resolve().parent / "_static/switcher.json"
    version_switcher_file.write_text(version_switcher)
    print(
        f"Version switcher written to {version_switcher_file}. "
        "Please check its contents and commit it."
    )

    # Create the release notes skeleton and write it to a new file
    release_notes = create_release_notes(
        formatted_release_version=formatted_release_version,
        formatted_release_date=formatted_release_date,
    )
    release_notes_file = (
        Path(__file__).resolve().parent
        / f"release_notes/{formatted_release_version}.md"
    )
    release_notes_file.write_text(release_notes)
    print(
        f"Release notes skeleton written to {release_notes_file}. "
        "Please fill in the release notes and commit it."
    )



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import logging
from pathlib import Path

import pytest

from cotainr.container import SingularitySandbox
from cotainr.tracing import LogDispatcher, LogSettings

from ..util.patches import patch_disable_stream_subprocess
from .data import data_cached_alpine_sif
from .patches import patch_fake_singularity_sandbox_env_folder


class TestConstructor:
    def test_attributes(self):
        base_image = "my_base_image_6021"
        sandbox = SingularitySandbox(base_image=base_image)
        assert sandbox.base_image == base_image
        assert sandbox.sandbox_dir is None
        assert sandbox.log_dispatcher is None
        assert sandbox._verbosity == 0
        assert sandbox.architecture is None

    def test_setup_log_dispatcher(self):
        log_settings = LogSettings(verbosity=1)
        sandbox = SingularitySandbox(
            base_image="my_base_image_6021", log_settings=log_settings
        )
        assert sandbox._verbosity == 1
        assert sandbox.log_dispatcher.verbosity == log_settings.verbosity
        assert sandbox.log_dispatcher.log_file_path == log_settings.log_file_path
        assert sandbox.log_dispatcher.no_color == log_settings.no_color
        assert sandbox.log_dispatcher.map_log_level is SingularitySandbox._map_log_level
        assert sandbox.log_dispatcher.logger_stdout.name == "SingularitySandbox.out"
        assert sandbox.log_dispatcher.logger_stderr.name == "SingularitySandbox.err"


class TestContext:
    def test_add_verbosity_arg(self, capsys, patch_disable_stream_subprocess):
        sandbox = SingularitySandbox(base_image="my_base_image_6021")
        sandbox.architecture = "test"
        with sandbox:
            pass
        stdout_lines = capsys.readouterr().out.rstrip("\n").split("\n")
        assert "args=['singularity', '-q', " in stdout_lines[0]

    @pytest.mark.singularity_integration
    def test_singularity_sandbox_creation(self, data_cached_alpine_sif):
        with SingularitySandbox(base_image=data_cached_alpine_sif) as sandbox:
            # Check that the sandbox contains the expected singularity files
            assert (sandbox.sandbox_dir / "environment").exists()
            assert (sandbox.sandbox_dir / "singularity").exists()
            assert (sandbox.sandbox_dir / ".singularity.d").exists()

    def test_tmp_dir_setup_and_teardown(self, patch_disable_stream_subprocess):
        test_dir = Path().resolve()
        sandbox = SingularitySandbox(base_image="my_base_image_6021")
        sandbox.architecture = "test"
        with sandbox:
            # Check that we are in the temporary sandbox directory
            sandbox_dir = sandbox.sandbox_dir
            assert sandbox_dir.stem == "singularity_sandbox"
            assert sandbox_dir != test_dir
            assert Path().resolve() == sandbox_dir

        # Check that we have exited and removed the temporary sandbox directory
        assert Path().resolve() == test_dir
        assert sandbox.sandbox_dir is None
        assert not sandbox_dir.exists()


class TestAddToEnv:
    def test_add_twice(
        self,
        patch_disable_stream_subprocess,
        patch_fake_singularity_sandbox_env_folder,
    ):
        lines = ["first script line", "second script line"]
        sandbox = SingularitySandbox(base_image="my_base_image_6021")
        sandbox.architecture = "test"
        with sandbox:
            env_file = sandbox.sandbox_dir / ".singularity.d/env/92-cotainr-env.sh"
            for line in lines:
                sandbox.add_to_env(shell_script=line)
            assert env_file.read_text() == lines[0] + "\n" + lines[1] + "\n"

    @pytest.mark.singularity_integration
    @pytest.mark.parametrize(
        "umask",
        [
            # umask  file permissions  directory permissions
            0o000,  # 666 (rw-rw-rw-)   777 (rwxrwxrwx)
            0o002,  # 664 (rw-rw-r--)   775 (rwxrwxr-x)
            0o007,  # 660 (rw-rw----)   770 (rwxrwx---)  # umask used on LUMI
            0o022,  # 644 (rw-r--r--)   755 (rwxr-xr-x)  # default umask in most cases
            0o027,  # 640 (rw-r-----)   750 (rwxr-x---)
            0o077,  # 600 (rw-------)   700 (rwx------)
        ],
    )
    def test_correct_umask(
        self,
        umask,
        context_set_umask,
        data_cached_alpine_sif,
        singularity_exec,
        tmp_path,
    ):
        with context_set_umask(umask):  # default umask on LUMI
            tmp_path.chmod(
                # apply effective umask permissions to tmp_path as well
                # it is created without the custom umask
                0o777 - umask
            )
            built_image_path = tmp_path / "built_image_6021"
            with SingularitySandbox(base_image=data_cached_alpine_sif) as sandbox:
                # Test file permissions
                env_file = sandbox.sandbox_dir / ".singularity.d/env/92-cotainr-env.sh"
                sandbox._create_file(f=env_file)
                assert env_file.exists()
                test_file_mode = env_file.stat().st_mode
                # file permissions extracted from the last 3 octal digits of st_mode
                test_file_permissions = test_file_mode & 0o777
                assert oct(test_file_permissions) == "0o644"

                # Test source 92-cotainr-env.sh
                sandbox.add_to_env(
                    shell_script="echo 'we can read the env file, 6021!'"
                )
                sandbox.build_image(path=built_image_path)

            assert singularity_exec(
                f"{built_image_path} echo 'more text...'"
            ).stdout == ("we can read the env file, 6021!\nmore text...\n")

    def test_newline_encapsulation(
        self,
        patch_disable_stream_subprocess,
        patch_fake_singularity_sandbox_env_folder,
    ):
        sandbox = SingularitySandbox(base_image="my_base_image_6021")
        sandbox.architecture = "test"
        with sandbox:
            env_file = sandbox.sandbox_dir / ".singularity.d/env/92-cotainr-env.sh"
            shell_script = "fancy shell_script\nas a double line string"
            sandbox.add_to_env(shell_script=shell_script)
            assert env_file.read_text() == shell_script + "\n"

    @pytest.mark.singularity_integration
    def test_existing_file(self, data_cached_alpine_sif):
        with SingularitySandbox(base_image=data_cached_alpine_sif) as sandbox:
            env_file = sandbox.sandbox_dir / ".singularity.d/env/92-cotainr-env.sh"
            existing_shell_script = "some existing\nshell script"

            # .write_text() creates the file with permissions corresponding to system default
            env_file.write_text(existing_shell_script)
            new_shell_script = "fancy_shell_script\nas_a_string"
            sandbox.add_to_env(shell_script=new_shell_script)
            assert (
                env_file.read_text().strip() == existing_shell_script + new_shell_script
            )

    @pytest.mark.singularity_integration
    def test_when_architecture_is_set(self, data_cached_alpine_sif):
        sandbox = SingularitySandbox(base_image=data_cached_alpine_sif)
        sandbox.architecture = "test"
        with sandbox:
            assert sandbox.architecture == "test"


@pytest.mark.singularity_integration
class TestBuildImage:
    def test_add_verbosity_arg(self, capsys, patch_disable_stream_subprocess):
        sandbox = SingularitySandbox(base_image="my_base_image_6021")
        sandbox.architecture = "test"
        with sandbox:
            sandbox.build_image(path="some_path_6021")
        stdout_lines = capsys.readouterr().out.rstrip("\n").split("\n")
        assert "args=['singularity', '-q', " in stdout_lines[-1]

    def test_environment_not_overwritten(
        self, data_cached_alpine_sif, singularity_exec, tmp_path
    ):
        # Test that any custom environment variables set via
        # SingularitySandbox.add_to_env(...) are not overwritten when the SIF
        # image file is built.
        # We used to write our custom environment variables to /environment
        # which is a symlink to /.singularity.d/env/90-environment.sh However,
        # as of some newer version of Apptainer, the content of /environment is
        # set to its default value when building the SIF image from the
        # sandbox, erasing our modifications. It is unclear if this is
        # intentional. Looking at both the Apptainer documentation
        # (https://apptainer.org/docs/user/main/environment_and_metadata.html#singularity-d-directory)
        # as well as the Singularity documentation
        # (https://docs.sylabs.io/guides/latest/user-guide/environment_and_metadata.html#singularity-d-directory)
        # it is sketchy to edit this /environment file as they both state that
        # "You should not manually modify files under /.singularity.d" and "In
        # the longer term, metadata will be moved outside of the container, and
        # stored only in the SIF file metadata descriptor."
        # As a workaround, we now write our custom environment variables to a
        # custom /.singularity.d/env/92-cotainr-env.sh file which is probably
        # not the right solution. However, at this point, though, it is unclear
        # how to set custom environment variables in any other way when using a
        # Singularity sandbox.
        build_container_path = tmp_path / "container_6021.sif"
        with SingularitySandbox(base_image=data_cached_alpine_sif) as sandbox:
            sandbox.add_to_env(shell_script="some shell script")
            assert (
                sandbox.sandbox_dir / ".singularity.d/env/92-cotainr-env.sh"
            ).exists()
            assert (
                (sandbox.sandbox_dir / ".singularity.d/env/92-cotainr-env.sh")
                .read_text()
                .endswith("some shell script\n")
            )
            sandbox.build_image(path=build_container_path)

        container_cat_env_process = singularity_exec(
            f"{build_container_path} cat /.singularity.d/env/92-cotainr-env.sh"
        )
        env_file_contents = container_cat_env_process.stdout
        assert env_file_contents.endswith("some shell script\n")

    def test_fix_perms_on_oci_docker_images(self, tmp_path):
        # Tests correct permission handling in relation to the error:
        #   FATAL:   While performing build: packer failed to pack: copy Failed:
        #   symlink GlobalSign_Root_R46.pem
        #   .../var/lib/ca-certificates/openssl/002c0b4f.0: permission denied
        # which seems to be a problem with all SUSE based docker images.
        # https://github.com/DeiC-HPC/cotainr/issues/48
        image_path = tmp_path / "image_6021"
        assert not image_path.exists()
        with SingularitySandbox(base_image="docker://opensuse/leap:15.4") as sandbox:
            sandbox.build_image(path=image_path)

        assert image_path.exists()

    def test_overwrite_existing_image(self, data_cached_alpine_sif, tmp_path):
        existing_singularity_image_path = tmp_path / "existing_image_6021"
        existing_singularity_image_path.write_bytes(data_cached_alpine_sif.read_bytes())
        existing_singularity_image_stats = existing_singularity_image_path.stat()
        assert existing_singularity_image_path.exists()
        with SingularitySandbox(base_image=data_cached_alpine_sif) as sandbox:
            sandbox.build_image(path=existing_singularity_image_path)

        # Check that the image file has been overwritten
        assert (
            existing_singularity_image_path.stat() != existing_singularity_image_stats
        )

    def test_sandbox_image_equality(
        self, data_cached_alpine_sif, singularity_exec, tmp_path
    ):
        built_image_path = tmp_path / "built_image_6021"
        with SingularitySandbox(base_image=data_cached_alpine_sif) as sandbox:
            sandbox.build_image(path=built_image_path)

        # Simple test of equality based on OS release of base image and built container
        base_image_os_release = singularity_exec(
            f"{data_cached_alpine_sif} cat /etc/os-release"
        ).stdout
        built_image_os_release = singularity_exec(
            f"{built_image_path} cat /etc/os-release"
        ).stdout
        assert base_image_os_release == built_image_os_release


@pytest.mark.singularity_integration
class TestRunCommandInContainer:
    def test_add_verbosity_arg(self, capsys, patch_disable_stream_subprocess):
        sandbox = SingularitySandbox(base_image="my_base_image_6021")
        sandbox.architecture = "test"
        with sandbox:
            sandbox.run_command_in_container(cmd="ls")
        stdout_lines = capsys.readouterr().out.rstrip("\n").split("\n")
        assert "args=['singularity', '-q', " in stdout_lines[-1]

    def test_correct_umask(self, data_cached_alpine_sif, context_set_umask):
        test_file = "test_file_6021"
        with context_set_umask(0o007):  # default umask on LUMI
            with SingularitySandbox(base_image=data_cached_alpine_sif) as sandbox:
                sandbox.run_command_in_container(cmd=f"touch /{test_file}")
                test_file_path = sandbox.sandbox_dir / test_file
                assert test_file_path.exists()
                test_file_mode = test_file_path.stat().st_mode
                # file permissions extracted from the last 3 octal digits of st_mode
                test_file_permissions = test_file_mode & 0o777
                assert oct(test_file_permissions) == "0o644"

    def test_error_handling(self, data_cached_alpine_sif):
        cmd = "some6021 non-meaningful command"
        with SingularitySandbox(base_image=data_cached_alpine_sif) as sandbox:
            with pytest.raises(
                ValueError, match=f"^Invalid command {cmd=} passed to Singularity"
            ):
                sandbox.run_command_in_container(cmd=cmd)

    def test_no_home(self, data_cached_alpine_sif):
        with SingularitySandbox(base_image=data_cached_alpine_sif) as sandbox:
            process = sandbox.run_command_in_container(cmd="ls -l /home")
        assert process.stdout.strip() == "total 0"


class Test_AssertWithinSandboxContext:
    def test_pass_inside_sandbox(self):
        sandbox = SingularitySandbox(base_image="my_base_image_6021")
        sandbox.sandbox_dir = Path()
        sandbox._assert_within_sandbox_context()

    def test_fail_outside_sandbox(self):
        sandbox = SingularitySandbox(base_image="my_base_image_6021")
        with pytest.raises(
            ValueError,
            match=r"^The operation is only valid inside a sandbox context\.$",
        ):
            sandbox._assert_within_sandbox_context()


class Test_CreateFile:
    def test_broken_subprocess(self, patch_disable_stream_subprocess):
        sandbox = SingularitySandbox(base_image="my_base_image_6021")
        sandbox.architecture = "test"
        with sandbox:
            any_file = sandbox.sandbox_dir / "anyfile.txt"
            with pytest.raises(FileNotFoundError):
                sandbox._create_file(f=any_file)


class Test_SubprocessRunner:
    def test_logger_prefix_for_custom_log_dispatcher(
        self,
        capsys,
        patch_disable_stream_subprocess,
    ):
        log_dispatcher = LogDispatcher(
            name="test_dispatcher_6021",
            map_log_level_func=lambda msg: logging.WARNING,
            log_settings=LogSettings(),
        )
        sandbox = SingularitySandbox(base_image="my_base_image_6021")
        sandbox._subprocess_runner(
            custom_log_dispatcher=log_dispatcher, args=["some_arg_6021"]
        )
        stderr = capsys.readouterr().err
        assert stderr.startswith("SingularitySandbox/test_dispatcher_6021.err")

    def test_use_own_log_dispatcher(self, capsys, patch_disable_stream_subprocess):
        sandbox = SingularitySandbox(
            base_image="my_base_image_6021", log_settings=LogSettings(verbosity=1)
        )
        sandbox._subprocess_runner(args=["some_arg_6021"])
        stderr = capsys.readouterr().err
        assert stderr.startswith("SingularitySandbox.err")


class Test_AddVerbosityArg:
    @pytest.mark.parametrize(
        ["verbosity", "verbosity_arg"],
        [
            (-1, "-s"),
            (0, "-q"),
            (1, None),
            (2, None),
            (3, "-v"),
            (4, "-d"),
            (5, "-d"),
            (6, "-d"),
        ],
    )
    def test_correct_mapping_of_verbosity(self, verbosity, verbosity_arg):
        sandbox = SingularitySandbox(base_image="my_base_image_6021")
        sandbox._verbosity = verbosity
        args = ["first_arg", "last_arg"]
        sandbox._add_verbosity_arg(args=args)
        if verbosity_arg is not None:
            assert len(args) == 3
            assert args[0] == "first_arg"
            assert args[1] == verbosity_arg
            assert args[2] == "last_arg"
        else:
            assert len(args) == 2
            assert args[0] == "first_arg"
            assert args[1] == "last_arg"

    def test_return_args(self):
        sandbox = SingularitySandbox(base_image="my_base_image_6021")
        args = ["some_arg"]
        returned_args = sandbox._add_verbosity_arg(args=args)
        assert returned_args is args


class Test_MapLogLevel:
    @pytest.mark.parametrize(
        ["msg", "log_level"],
        [
            ("DEBUG", logging.DEBUG),
            ("VERBOSE", logging.DEBUG),
            ("DEBUG:6021", logging.DEBUG),
            ("VERBOSE 6021", logging.DEBUG),
            ("INFO", logging.INFO),
            ("LOG", logging.INFO),
            ("INFORMATION", logging.INFO),
            ("LOG some text", logging.INFO),
            ("WARNING", logging.WARNING),
            ("WARNING-log", logging.WARNING),
            ("ERROR", logging.ERROR),
            ("ERROR_MSG", logging.ERROR),
            ("ABRT", logging.CRITICAL),
            ("FATAL", logging.CRITICAL),
            ("ABRT/mission", logging.CRITICAL),
            ("FATALities", logging.CRITICAL),
            ("unknown", logging.INFO),
            (
                "some messages containing DEBUG, VERBOSE, WARNING, ERROR, ABRT, and FATAL",
                logging.INFO,
            ),
        ],
    )
    def test_correct_log_level_mapping(self, msg, log_level):
        assert SingularitySandbox._map_log_level(msg=msg) == log_level



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk Licensed under the European Union Public License (EUPL)
1.2
 - see the LICENSE file for details.

This module implements dynamic version number computation for cotainr based on
git history (if in a git development environment) or the installed package
version (if used from an installed cotainr package). If neither is available,
the reported version is "<unknown version>".

Adapted from the hatch-vcs-footgun-example, see
<https://github.com/maresb/hatch-vcs-footgun-example/>.

Functions
---------
_get_cotainr_calver_tag_pattern()
    Load the cotainr calver tag pattern from pyproject.toml.
_get_hatch_version()
    Compute the version number in a development environment.
_get_importlib_metadata_version()
    Get the version number for an installed cotainr package.

Attributes
----------
__version__
    The cotainr version number.
"""

from pathlib import Path


def _get_cotainr_calver_tag_pattern():
    """
    Load the cotainr calver tag pattern from pyproject.toml.

    The hatch version tag pattern is the single source of truth for the regex
    defining the calver pattern for cotainr version numbers.

    Returns
    -------
    cotainr_calver_tag_pattern : str
        The cotainr calver tag pattern regex.
    """
    pyproject_toml_path = Path(__file__).resolve().parents[1] / "pyproject.toml"

    # MARK_PYTHON_VERSION: Update this code when we stop supporting Python < 3.11.
    # The tomllib module is only available in Python 3.11 and later.
    # Until we only support Python 3.11 and later, we use this custom code to
    # read the pyproject.toml file. It attempts to do the same thing as:
    #   import tomllib
    #   cotainr_calver_tag_pattern = tomllib.loads(pyproject_toml_path.read_text())["tool"]["hatch"]["version"]["tag-pattern"]
    with open(pyproject_toml_path) as f:
        for line in f:
            if line.startswith("tag-pattern"):
                # Split line on "=" and remove #comments, whitespace, and single quotes
                cotainr_calver_tag_pattern = (
                    line.split("=")[1].split("#")[0].strip().strip("'")
                )
                break
        else:
            raise RuntimeError(
                "Unable to find the cotainr calver tag pattern in pyproject.toml"
            )

    return cotainr_calver_tag_pattern


def _get_hatch_version():
    """
    Compute the version number in a development environment.

    Uses the hatchling package functionality to determine the version number.

    Returns
    -------
    str or None
        The version number or `None` if hatchling is unable to determine the
        version number.
    """
    try:
        from hatchling.metadata.core import ProjectMetadata
        from hatchling.plugin.manager import PluginManager
    except ImportError:
        return None

    file_path = Path(__file__)
    if file_path.parent.stem == "site-packages":
        # If the parent directory is "site-packages", it looks like we are in
        # an environment with several installed Python packages. No
        # pyproject.toml file should be available in "site-packages", but if it
        # is, we can't be sure which package it belongs to, so don't use hatch
        # to determine the version number.
        return None
    else:
        metadata = ProjectMetadata(
            root=f"{file_path.parents[1]}", plugin_manager=PluginManager()
        )

    try:
        vcs_version = metadata.hatch.version.cached
    except Exception:
        # If "cached" doesn't exist or raises an exception, give up and return None.
        return None

    assert isinstance(vcs_version, str)
    return vcs_version


def _get_importlib_metadata_version():
    """
    Get the version number for an installed cotainr package.

    Uses standard importlib.metadata functions to get the version number for an
    installed package.

    Returns
    -------
    str or None
        The version number or `None` if not used from an installed package.

    """
    from importlib.metadata import PackageNotFoundError, version

    try:
        package_version = version(__package__)
    except PackageNotFoundError:
        return None

    assert isinstance(package_version, str)
    return package_version


# Priority queue: First hatch version, then import lib or else the default value
__version__ = (
    _get_hatch_version() or _get_importlib_metadata_version() or "unknown version"
)



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import sys

from ._version import __version__

__all__ = ["__version__"]

_minimum_dependency_version = {
    # Versions must be specified as a (major, minor, patchlevel) tuple of
    # integers
    "python": (  # MARK_PYTHON_VERSION: Update this to reflect the minimum supported Python version.
        3,
        9,
        0,
    ),
    "apptainer": (  # MARK_APPTAINER_VERSION: Update this to reflect the minimum supported Apptainer version.
        1,
        3,
        4,
    ),
    "singularity-ce": (  # MARK_APPTAINER_VERSION: Update this to reflect the minimum supported SingularityCE version.
        3,
        9,
        2,
    ),
}

# Error early on too old Python version
if sys.version_info < _minimum_dependency_version["python"]:  # pragma: no cover
    # Note that this is using percent formatting on purpose, so this
    # file can be executed by Python versions older than the minimum
    # for Cotainr in general; this includes being able to run with
    # Python 2.7!
    sys.exit(
        (  # noqa: UP031
            "\x1b[38;5;160m"  # start of red colored text
            "Cotainr requires Python>=%s\n"
            "You are running Python==%s\n"
            "from '%s'\n"
            "ABORTING!"
            "\x1b[0m"  # end of red colored text
        )
        % (
            ".".join(map(str, _minimum_dependency_version["python"])),
            sys.version,
            sys.executable,
        )
    )



"""
cotainr - a user space Apptainer/Singularity container builder.

Copyright DeiC, deic.dk
Licensed under the European Union Public License (EUPL) 1.2
- see the LICENSE file for details.

"""

import argparse
from pathlib import Path
import re
import shlex

import pytest

from cotainr.cli import Build, CotainrCLI

from ..container.patches import (
    patch_disable_add_metadata,
    patch_disable_singularity_sandbox_subprocess_runner,
    patch_fake_singularity_sandbox_env_folder,
    patch_save_singularity_sandbox_context,
)
from ..pack.patches import (
    patch_disable_conda_install_bootstrap_conda,
    patch_disable_conda_install_display_miniforge_license_for_acceptance,
    patch_disable_conda_install_download_miniforge_installer,
)
from ..tracing.patches import patch_disable_console_spinner
from ..util.patches import patch_empty_system, patch_system_with_actual_file


class TestConstructor:
    def test_nonexisting_conda_env(self):
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        conda_env = "some_conda_env_6021"
        with pytest.raises(FileNotFoundError) as exc_info:
            Build(image_path=image_path, base_image=base_image, conda_env=conda_env)
        exc_msg = str(exc_info.value)
        assert exc_msg.startswith("The provided Conda env file '")
        assert exc_msg.endswith("' does not exist.")
        assert conda_env in exc_msg

    @pytest.mark.parametrize(
        "base_image,system",
        [("some_base_image_6021", None), (None, "some_system_6021")],
    )
    def test_only_specifying_required_args(
        self, base_image, system, patch_system_with_actual_file
    ):
        # See also the matching TestAddArguments test below
        image_path = "some_image_path_6021"
        build = Build(image_path=image_path, base_image=base_image, system=system)
        assert build.image_path.is_absolute()
        assert build.image_path.name == image_path
        if system is not None:
            base_image = "some_base_image_6021"
            assert build.system["base-image"] == base_image
            assert build.base_image == base_image
        else:
            assert build.base_image == base_image
        assert build.conda_env is None
        assert not build.accept_licenses
        assert build.log_settings.verbosity == 0
        assert build.log_settings.log_file_path is None
        assert not build.log_settings.no_color

    @pytest.mark.parametrize(
        "base_image,system",
        [("some_base_image_6021", None), (None, "some_system_6021")],
    )
    def test_specifying_conda_env(
        self, base_image, system, patch_system_with_actual_file
    ):
        # See also the matching TestAddArguments test below
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        conda_env = "some_conda_env_6021"
        Path(conda_env).touch()
        build = Build(
            image_path=image_path,
            base_image=base_image,
            system=system,
            conda_env=conda_env,
        )
        assert build.conda_env.is_absolute()
        assert build.conda_env.name == conda_env

    def test_specifying_non_existing_system(self, patch_empty_system):
        image_path = "some_image_path_6021"
        system = "some_system_6021"
        with pytest.raises(KeyError, match="System does not exist"):
            Build(image_path=image_path, system=system)

    @pytest.mark.parametrize("answer", ["no", "No", "NO"])
    def test_already_existing_file_but_no(
        self, answer, factory_mock_input, monkeypatch
    ):
        monkeypatch.setattr("builtins.input", factory_mock_input(answer))
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        Path(image_path).touch()
        with pytest.raises(SystemExit):
            Build(image_path=image_path, base_image=base_image)

    @pytest.mark.parametrize("answer", ["yes", "Yes", "YES"])
    def test_already_existing_file(self, answer, factory_mock_input, monkeypatch):
        monkeypatch.setattr("builtins.input", factory_mock_input(answer))
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        Path(image_path).touch()
        build = Build(image_path=image_path, base_image=base_image)
        assert build.base_image == base_image

    @pytest.mark.parametrize(
        "answer_sequence", [["blahblah", "yes"], ["1", "2", "", "True", "False", "yes"]]
    )
    def test_already_existing_file_initial_invalid_answer(
        self, answer_sequence, factory_mock_input_sequence, monkeypatch, capsys
    ):
        monkeypatch.setattr(
            "builtins.input", factory_mock_input_sequence(answer_sequence)
        )
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        Path(image_path).touch()
        build = Build(image_path=image_path, base_image=base_image)
        stdout = capsys.readouterr().out
        assert build.base_image == base_image
        assert all(  # Should print error message about invalid input for each invalid answer
            answer in stdout for answer in answer_sequence[:-1]
        )

    def test_specifying_accept_licenses(self):
        # See also the matching TestAddArguments test below
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        build = Build(
            image_path=image_path, base_image=base_image, accept_licenses=True
        )
        assert build.accept_licenses

    def test_specifying_log_to_file(self):
        # See also the matching TestAddArguments test below
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        build = Build(image_path=image_path, base_image=base_image, log_to_file=True)
        print(
            build.log_settings.log_file_path.name,
        )
        assert re.match(
            r"^cotainr_build_\d{4}-\d{2}-\d{2}T\d{2}\.\d{2}\.\d{2}\.\d{6}$",
            build.log_settings.log_file_path.name,
        )

    def test_specifying_no_color(self):
        # See also the matching TestAddArguments test below
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        build = Build(image_path=image_path, base_image=base_image, no_color=True)
        assert build.log_settings.no_color


class TestAddArguments:
    def test_not_specifying_base_image_or_system(self, capsys):
        # See also the matching TestConstructor test above
        parser = argparse.ArgumentParser(prog="some_prog_name_6021")
        Build.add_arguments(parser=parser)
        with pytest.raises(SystemExit):
            parser.parse_args(args=shlex.split("some_image_path_6021"))
        stderr = capsys.readouterr().err
        assert stderr.strip().endswith(
            "some_prog_name_6021: error: one of the arguments --base-image --system is required"
        )

    def test_only_specifying_required_args_base_image(self):
        # See also the matching TestConstructor test above
        parser = argparse.ArgumentParser()
        Build.add_arguments(parser=parser)
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        args = parser.parse_args(
            args=shlex.split(f"{image_path} --base-image={base_image}")
        )
        assert isinstance(args.image_path, Path)
        assert args.image_path.name == image_path
        assert isinstance(args.base_image, str)
        assert args.base_image == base_image
        assert not args.accept_licenses

    def test_only_specifying_required_args_system(self, patch_system_with_actual_file):
        # See also the matching TestConstructor test above
        parser = argparse.ArgumentParser()
        Build.add_arguments(parser=parser)
        image_path = "some_image_path_6021"
        system = "some_system_6021"
        args = parser.parse_args(args=shlex.split(f"{image_path} --system={system}"))
        assert isinstance(args.image_path, Path)
        assert args.image_path.name == image_path
        assert isinstance(args.system, str)
        assert args.system == system
        assert not args.accept_licenses

    def test_specifying_both_system_and_base_image(self, capsys):
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        system = "some_system_6021"
        parser = argparse.ArgumentParser()
        Build.add_arguments(parser=parser)

        with pytest.raises(SystemExit):
            parser.parse_args(
                args=shlex.split(
                    f"{image_path} --system {system} --base-image {base_image}"
                )
            )
        stderr = capsys.readouterr().err
        assert stderr.strip().endswith(
            "argument --base-image: not allowed with argument --system"
        )

    def test_specifying_conda_env_arg(self):
        # See also the matching TestConstructor test above
        parser = argparse.ArgumentParser()
        Build.add_arguments(parser=parser)
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        conda_env = "some_conda_env_6021"
        args = parser.parse_args(
            args=shlex.split(
                f"{image_path} --base-image={base_image} --conda-env={conda_env}"
            )
        )
        assert isinstance(args.conda_env, Path)
        assert args.conda_env.name == conda_env

    def test_specifying_accept_licenses(self):
        # See also the matching TestConstructor test above
        parser = argparse.ArgumentParser()
        Build.add_arguments(parser=parser)
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        args = parser.parse_args(
            args=shlex.split(
                f"{image_path} --base-image={base_image} --accept-licenses"
            )
        )
        assert args.accept_licenses

    @pytest.mark.parametrize(
        ["verbose_arg", "verbosity"],
        [
            ("--verbose", 1),
            ("--verbose --verbose", 2),
            ("-v", 1),
            ("-vv", 2),
            ("-vvv", 3),
            ("-vvvv", 4),
        ],
    )
    def test_specifying_verbose(self, verbose_arg, verbosity):
        parser = argparse.ArgumentParser()
        Build.add_arguments(parser=parser)
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        args = parser.parse_args(
            args=shlex.split(f"{image_path} --base-image={base_image} {verbose_arg}")
        )
        assert args.verbosity == verbosity

    @pytest.mark.parametrize(
        ["verbose_arg", "verbosity"],
        [
            ("--quiet", -1),
            ("--quiet --quiet", -1),
            ("-q", -1),
            ("-qq", -1),
        ],
    )
    def test_specifying_quiet(self, verbose_arg, verbosity):
        parser = argparse.ArgumentParser()
        Build.add_arguments(parser=parser)
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        args = parser.parse_args(
            args=shlex.split(f"{image_path} --base-image={base_image} {verbose_arg}")
        )
        assert args.verbosity == verbosity

    def test_specifying_both_verbose_and_quiet(self, capsys):
        parser = argparse.ArgumentParser()
        Build.add_arguments(parser=parser)
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"

        with pytest.raises(SystemExit):
            parser.parse_args(
                args=shlex.split(
                    f"{image_path} --base-image={base_image} --verbose --quiet"
                )
            )
        stderr = capsys.readouterr().err
        assert stderr.strip().endswith(
            "argument --quiet/-q: not allowed with argument --verbose/-v"
        )

    def test_specifying_log_to_file(self):
        # See also the matching TestConstructor test above
        parser = argparse.ArgumentParser()
        Build.add_arguments(parser=parser)
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        args = parser.parse_args(
            args=shlex.split(f"{image_path} --base-image={base_image} --log-to-file")
        )
        assert args.log_to_file

    def test_specifying_no_color(self):
        # See also the matching TestConstructor test above
        parser = argparse.ArgumentParser()
        Build.add_arguments(parser=parser)
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        args = parser.parse_args(
            args=shlex.split(f"{image_path} --base-image={base_image} --no-color")
        )
        assert args.no_color


class TestExecute:
    def test_default_container_build(
        self,
        patch_disable_singularity_sandbox_subprocess_runner,
        # add_metadata fails as there is no sandbox_dir and labels.json since
        # we request patch_disable_singularity_sandbox_subprocess_runner.
        patch_disable_add_metadata,
        patch_disable_console_spinner,
        capsys,
    ):
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        Build(image_path=image_path, base_image=base_image).execute()
        (sandbox_create_cmd, sandbox_uname_cmd, sandbox_build_cmd) = (
            capsys.readouterr().out.strip().split("\n")
        )

        assert "'uname', '-m'" in sandbox_uname_cmd
        assert sandbox_create_cmd.startswith("PATCH: Ran command in sandbox:")
        assert all(
            s in sandbox_create_cmd
            for s in ["'singularity'", "'build'", "'--sandbox'", f"'{base_image}'"]
        )
        assert sandbox_build_cmd.startswith("PATCH: Ran command in sandbox:")
        assert all(
            s in sandbox_build_cmd
            for s in ["'singularity'", "'build'", f"{image_path}"]
        )

    def test_include_conda_env(
        self,
        patch_disable_singularity_sandbox_subprocess_runner,
        patch_disable_conda_install_bootstrap_conda,
        patch_disable_conda_install_download_miniforge_installer,
        patch_fake_singularity_sandbox_env_folder,
        patch_save_singularity_sandbox_context,
        patch_disable_add_metadata,
        patch_disable_console_spinner,
        caplog,
        capsys,
    ):
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        conda_env = "some_conda_env_6021"
        conda_env_content = "Some conda env content 6021"
        saved_sandbox_dir = Path(f"./{patch_save_singularity_sandbox_context}")
        Path(conda_env).write_text(conda_env_content)
        Build(
            image_path=image_path,
            base_image=base_image,
            conda_env=conda_env,
            accept_licenses=True,
        ).execute()

        # Check that conda_env file has been copied to container
        assert (saved_sandbox_dir / f"{conda_env}").read_text() == conda_env_content

        # Check that the singularity environment has been updated activate conda env
        assert (
            (saved_sandbox_dir / ".singularity.d/env/92-cotainr-env.sh")
            .read_text()
            .strip()
            .endswith("conda activate conda_container_env")
        )

        # Check sandbox interaction commands
        (
            sandbox_create_cmd,
            sandbox_uname_cmd,
            conda_bootstrap_cmd,
            conda_bootstrap_clean_cmd,
            conda_env_create_cmd,
            conda_clean_cmd,
            sandbox_build_cmd,
        ) = capsys.readouterr().out.strip().split("\n")
        assert sandbox_create_cmd.startswith("PATCH: Ran command in sandbox:")
        assert "'uname', '-m'" in sandbox_uname_cmd
        assert all(
            s in sandbox_create_cmd
            for s in ["'singularity'", "'build'", "'--sandbox'", f"'{base_image}'"]
        )
        assert conda_bootstrap_cmd.startswith("PATCH: Bootstrapped Conda using")
        assert conda_bootstrap_clean_cmd.startswith("PATCH: Ran command in sandbox:")
        assert "'conda', 'clean'" in conda_bootstrap_clean_cmd
        assert conda_env_create_cmd.startswith("PATCH: Ran command in sandbox:")
        assert all(
            s in conda_env_create_cmd
            for s in [
                "'conda'",
                "'env'",
                "'create'",
                f"{conda_env}",
                "'conda_container_env'",
            ]
        )
        assert conda_clean_cmd.startswith("PATCH: Ran command in sandbox:")
        assert "'conda', 'clean'" in conda_clean_cmd
        assert sandbox_build_cmd.startswith("PATCH: Ran command in sandbox:")
        assert all(
            s in sandbox_build_cmd
            for s in ["'singularity'", "'build'", f"{image_path}"]
        )

        # Check log calls
        assert re.search(
            r"^WARNING  CondaInstall\.err\:pack\.py\:(\d+) "
            r"You have accepted the Miniforge installer license via the command line option "
            r"'--accept-licenses'\.$",
            caplog.text,
            flags=re.MULTILINE,
        )

    def test_no_beforehand_license_acceptance(
        self,
        patch_disable_singularity_sandbox_subprocess_runner,
        patch_disable_conda_install_download_miniforge_installer,
        patch_disable_conda_install_display_miniforge_license_for_acceptance,
    ):
        image_path = "some_image_path_6021"
        base_image = "some_base_image_6021"
        conda_env = "some_conda_env_6021"
        conda_env_content = "Some conda env content 6021"
        Path(conda_env).write_text(conda_env_content)
        with pytest.raises(SystemExit, match="^PATCH: Showing license terms for "):
            Build(
                image_path=image_path,
                base_image=base_image,
                conda_env=conda_env,
                accept_licenses=False,
            ).execute()


class TestHelpMessage:
    def test_CLI_subcommand_help_message(self, argparse_options_line, capsys):
        with pytest.raises(SystemExit):
            CotainrCLI(args=["build", "--help"])
        stdout = capsys.readouterr().out
        # whitespaces in the formatting often leads to failed tests.
        # especially true when doing this test for Python =< 3.12 and >=3.13 due to differences in whitespace trimming.
        # thus remove all \\n \\r \\t \\f and spaces.
        stdout = " ".join(stdout.split())
        target = (
            # Capsys apparently assumes an 80 char terminal (?) - thus extra '\n'
            "usage: cotainr build [-h] (--base-image BASE_IMAGE | --system SYSTEM)\n"
            "                     [--conda-env CONDA_ENV] [--accept-licenses]\n"
            "                     [--verbose | --quiet] [--log-to-file] [--no-color]\n"
            "                     image_path\n\n"
            "Build a container.\n\n"
            "positional arguments:\n"
            "  image_path            path to the built container image\n\n"
            f"{argparse_options_line}"
            "  -h, --help            show this help message and exit\n"
            "  --base-image BASE_IMAGE\n"
            "                        base image to use for the container which may be any\n"
            "                        valid Apptainer/Singularity <BUILD SPEC>\n"
            "  --system SYSTEM       which system/partition you will be running the\n"
            "                        container on. This sets base image and other\n"
            "                        parameters for a simpler container creation. Running\n"
            "                        the info command will tell you more about the system\n"
            "                        and what is available\n"
            "  --conda-env CONDA_ENV\n"
            "                        path to a Conda environment.yml file to install and\n"
            "                        activate in the container. When installing a Conda\n"
            "                        environment, you must accept the Miniforge license\n"
            "                        terms, as specified during the build process\n"
            "  --accept-licenses     accept all license terms (if any) needed for\n"
            "                        completing the container build process\n"
            "  --verbose, -v         increase the verbosity of the output from cotainr. Can\n"
            "                        be used multiple times: Once for subprocess output,\n"
            "                        twice for subprocess INFO, three times for VERBOSE,\n"
            "                        four times for DEBUG and five times for TRACE.\n"
            "  --quiet, -q           do not show any non-CRITICAL output from cotainr\n"
            "  --log-to-file         create files containing all logging information shown\n"
            "                        on stdout/stderr\n"
            "  --no-color            do not use colored console output\n"
        )
        target = " ".join(target.split())
        assert target == stdout
